import logging
from pathlib import Path
from typing import Dict, Optional, Tuple

import pytest

from avtdl.core.monitors import load_headers


def add_to_file(text: str, path: Path, encoding: Optional[str] = None):
    with open(path, 'a+t', encoding=encoding) as fp:
        fp.write(text)


testcases: Dict[str, Tuple[str, str, Optional[Dict[str, str]]]] = {
    'valid_plaintext_file': (
        'headers.txt',
        'key1: value1\nkey2: value2\n',
        {'key1': 'value1', 'key2': 'value2'}
    ),
    'valid_json_file': (
        'headers.json',
        '{"key1": "value1", "key2": "value2"}',
        {'key1': 'value1', 'key2': 'value2'}
    ),
    'empty_plaintext_file': (
        'headers.txt',
        '',
        {}
    ),
    'empty_json_file': (
        'headers.json',
        '',
        None
    ),
    'invalid_json_file': (
        'headers.json',
        '{"key1": "value1", "key2": "value2",}',
        None
    ),
    'plaintext_file_empty_lines_ignored': (
        'headers.txt',
        '\nkey1: value1\n\nkey2: value2\n\n\n',
        {'key1': 'value1', 'key2': 'value2'}
    ),
    'plaintext_file_comments_ignored': (
        'headers.txt',
        'key1: value1\nkey2: value2\n# key3: value3',
        {'key1': 'value1', 'key2': 'value2'}
    ),
    'plaintext_file_only_invalid_value_skipped': (
        'headers.txt',
        'key1: value1\nkey2: value2\nkey3 value3',
        {'key1': 'value1', 'key2': 'value2'}
    ),
    'plaintext_file_no_space_after_semicolon': (
        'headers.txt',
        'key1:value1\nkey2:value2\n',
        {'key1': 'value1', 'key2': 'value2'}
    ),
    'plaintext_file_parentheses_preserved': (
        'headers.txt',
        '"key1": value1\nkey2: "value2"\n',
        {'"key1"': 'value1', 'key2': '"value2"'}
    ),
    'plaintext_file_multiple_semicolons': (
        'headers.txt',
        'key1: value1\nkey2: value: 2\n',
        {'key1': 'value1', 'key2': 'value: 2'}
    ),
}


@pytest.mark.parametrize('name, content, expected', testcases.values(), ids=testcases.keys())
def test_headers_file_parsing(tmp_path, name, content, expected):
    path = tmp_path / name
    add_to_file(content, path)
    result = load_headers(path, logging.getLogger('test_headers'))
    assert result == expected



# Prepare request to retrieve json data from graphql endpoints
#
# - timeline (home and chronological)
# - user tweets (without and with replies)
# - user likes
# - single tweet
#
# - user id by screen name
import abc
import datetime
import enum
import json
import logging
import urllib.parse
from typing import Any, Dict, Optional, Union

from multidict import CIMultiDictProxy

from avtdl.core.request import BucketRateLimit, Endpoint, HttpResponse, RequestDetails, get_retry_after
from avtdl.core.utils import find_one, get_cookie_value

TIMELINE_FEATURES = '{"rweb_video_screen_enabled":false,"payments_enabled":false,"profile_label_improvements_pcf_label_in_post_enabled":true,"responsive_web_profile_redirect_enabled":false,"rweb_tipjar_consumption_enabled":true,"verified_phone_label_enabled":false,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"premium_content_api_read_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"responsive_web_grok_analyze_button_fetch_trends_enabled":false,"responsive_web_grok_analyze_post_followups_enabled":true,"responsive_web_jetfuel_frame":true,"responsive_web_grok_share_attachment_enabled":true,"articles_preview_enabled":true,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"responsive_web_grok_show_grok_translated_post":false,"responsive_web_grok_analysis_button_from_backend":true,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_grok_image_annotation_enabled":true,"responsive_web_grok_imagine_annotation_enabled":true,"responsive_web_grok_community_note_auto_translation_is_enabled":false,"responsive_web_enhance_cards_enabled":false}'
USER_TWEETS_FEATURES = '{"rweb_video_screen_enabled":false,"payments_enabled":false,"profile_label_improvements_pcf_label_in_post_enabled":true,"responsive_web_profile_redirect_enabled":false,"rweb_tipjar_consumption_enabled":true,"verified_phone_label_enabled":false,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"premium_content_api_read_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"responsive_web_grok_analyze_button_fetch_trends_enabled":false,"responsive_web_grok_analyze_post_followups_enabled":true,"responsive_web_jetfuel_frame":true,"responsive_web_grok_share_attachment_enabled":true,"articles_preview_enabled":true,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"responsive_web_grok_show_grok_translated_post":false,"responsive_web_grok_analysis_button_from_backend":true,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_grok_image_annotation_enabled":true,"responsive_web_grok_imagine_annotation_enabled":true,"responsive_web_grok_community_note_auto_translation_is_enabled":false,"responsive_web_enhance_cards_enabled":false}fieldToggles={"withArticlePlainText":false}'
TWEET_DETAIL_FEATURES = '{"rweb_video_screen_enabled":false,"payments_enabled":false,"profile_label_improvements_pcf_label_in_post_enabled":true,"responsive_web_profile_redirect_enabled":false,"rweb_tipjar_consumption_enabled":true,"verified_phone_label_enabled":false,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"premium_content_api_read_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"responsive_web_grok_analyze_button_fetch_trends_enabled":false,"responsive_web_grok_analyze_post_followups_enabled":true,"responsive_web_jetfuel_frame":true,"responsive_web_grok_share_attachment_enabled":true,"articles_preview_enabled":true,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"responsive_web_grok_show_grok_translated_post":false,"responsive_web_grok_analysis_button_from_backend":true,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_grok_image_annotation_enabled":true,"responsive_web_grok_imagine_annotation_enabled":true,"responsive_web_grok_community_note_auto_translation_is_enabled":false,"responsive_web_enhance_cards_enabled":false}&fieldToggles={"withArticleRichContentState":true,"withArticlePlainText":false,"withGrokAnalyze":false,"withDisallowedReplyControls":false}'
USER_FEATURES = '{"hidden_profile_subscriptions_enabled":true,"payments_enabled":false,"profile_label_improvements_pcf_label_in_post_enabled":true,"responsive_web_profile_redirect_enabled":false,"rweb_tipjar_consumption_enabled":true,"verified_phone_label_enabled":false,"subscriptions_verification_info_is_identity_verified_enabled":true,"subscriptions_verification_info_verified_since_enabled":true,"highlights_tweets_tab_ui_enabled":true,"responsive_web_twitter_article_notes_tab_enabled":true,"subscriptions_feature_can_gift_premium":true,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"responsive_web_graphql_timeline_navigation_enabled":true}&fieldToggles={"withAuxiliaryUserLabels":true}'
SPACE_FEATURES = '{"spaces_2022_h2_spaces_communities":true,"spaces_2022_h2_clipping":true,"creator_subscriptions_tweet_preview_api_enabled":true,"payments_enabled":false,"profile_label_improvements_pcf_label_in_post_enabled":true,"responsive_web_profile_redirect_enabled":false,"rweb_tipjar_consumption_enabled":true,"verified_phone_label_enabled":false,"premium_content_api_read_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"responsive_web_grok_analyze_button_fetch_trends_enabled":false,"responsive_web_grok_analyze_post_followups_enabled":true,"responsive_web_jetfuel_frame":true,"responsive_web_grok_share_attachment_enabled":true,"articles_preview_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"responsive_web_grok_show_grok_translated_post":false,"responsive_web_grok_analysis_button_from_backend":true,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_grok_image_annotation_enabled":true,"responsive_web_grok_imagine_annotation_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_grok_community_note_auto_translation_is_enabled":false,"responsive_web_enhance_cards_enabled":false}'
SEARCH_FEATURES = '{"rweb_video_screen_enabled":false,"payments_enabled":false,"profile_label_improvements_pcf_label_in_post_enabled":true,"responsive_web_profile_redirect_enabled":false,"rweb_tipjar_consumption_enabled":true,"verified_phone_label_enabled":false,"creator_subscriptions_tweet_preview_api_enabled":true,"responsive_web_graphql_timeline_navigation_enabled":true,"responsive_web_graphql_skip_user_profile_image_extensions_enabled":false,"premium_content_api_read_enabled":false,"communities_web_enable_tweet_community_results_fetch":true,"c9s_tweet_anatomy_moderator_badge_enabled":true,"responsive_web_grok_analyze_button_fetch_trends_enabled":false,"responsive_web_grok_analyze_post_followups_enabled":true,"responsive_web_jetfuel_frame":true,"responsive_web_grok_share_attachment_enabled":true,"articles_preview_enabled":true,"responsive_web_edit_tweet_api_enabled":true,"graphql_is_translatable_rweb_tweet_is_translatable_enabled":true,"view_counts_everywhere_api_enabled":true,"longform_notetweets_consumption_enabled":true,"responsive_web_twitter_article_tweet_consumption_enabled":true,"tweet_awards_web_tipping_enabled":false,"responsive_web_grok_show_grok_translated_post":false,"responsive_web_grok_analysis_button_from_backend":true,"creator_subscriptions_quote_tweet_preview_enabled":false,"freedom_of_speech_not_reach_fetch_enabled":true,"standardized_nudges_misinfo":true,"tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled":true,"longform_notetweets_rich_text_read_enabled":true,"longform_notetweets_inline_media_enabled":true,"responsive_web_grok_image_annotation_enabled":true,"responsive_web_grok_imagine_annotation_enabled":true,"responsive_web_grok_community_note_auto_translation_is_enabled":false,"responsive_web_enhance_cards_enabled":false}'


class EndpointUrl:
    TIMELINE = 'https://twitter.com/i/api/graphql/vrNCYudm0qsExOy_-N9Q1g/HomeTimeline'
    LATEST_TIMELINE = 'https://twitter.com/i/api/graphql/rgMLKEHu1RpVEnuqlnwyiQ/HomeLatestTimeline'
    USER_TWEETS = 'https://twitter.com/i/api/graphql/oRJs8SLCRNRbQzuZG93_oA/UserTweets'
    USER_TWEETS_AND_REPLIES = 'https://twitter.com/i/api/graphql/kkaJ0Mf34PZVarrxzLihjg/UserTweetsAndReplies'
    TWEET_DETAIL = 'https://twitter.com/i/api/graphql/YVyS4SfwYW7Uw5qwy0mQCA/TweetDetail'
    USER_BY_SCREEN_NAME = 'https://twitter.com/i/api/graphql/ZHSN3WlvahPKVvUxVQbg1A/UserByScreenName'
    AUDIOSPACE_BY_ID = 'https://twitter.com/i/api/graphql/pCUWlI5FNL7ROBjmBsH3Zw/AudioSpaceById'
    SEARCH_TIMELINE = 'https://twitter.com/i/api/graphql/7r8ibjHuK3MWUyzkzHNMYQ/SearchTimeline'


def replace_url_host(url: str, new_host: str) -> str:
    parsed_host = urllib.parse.urlparse(new_host)
    parsed_url = urllib.parse.urlparse(url)
    scheme = parsed_url.scheme or parsed_host.scheme or 'https://'
    new_url = urllib.parse.urlunparse((scheme, parsed_host.netloc, parsed_url.path, parsed_url.params, parsed_url.query, parsed_url.fragment))
    return new_url


def get_netloc(host: str) -> str:
    return urllib.parse.urlparse(host).netloc


def get_auth_headers(cookies) -> dict[str, Any]:
    ct0 = get_cookie_value(cookies, 'ct0') or ''
    THE_API_KEY = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs=1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'
    headers = {
        'authorization': THE_API_KEY,
        'x-csrf-token': ct0
    }
    return headers


class TwitterRateLimit(BucketRateLimit):

    def _submit_headers(self, response: HttpResponse, logger: logging.Logger) -> bool:
        headers = response.headers
        try:
            # "default" argument is set to 'NaN' to trigger ValueError, since None is disliked by typechecker
            self.limit_total = int(headers.get('x-rate-limit-limit', 'NaN'))
            self.limit_remaining = int(headers.get('x-rate-limit-remaining', 'NaN'))
            self.reset_at = int(headers.get('x-rate-limit-reset', 'NaN'))
        except ValueError:
            logger.warning(f'[{self.name}] error parsing rate limit headers: "{headers}"')
            return False
        else:
            logger.debug(f'[{self.name}] rate limit {self.limit_remaining}/{self.limit_total}, resets after {datetime.timedelta(seconds=self.delay)}')
            return True


class TwitterEndpoint(Endpoint):
    """
    Base class providing common features for Twitter endpoints

    They might use methods provided by this class for convenience,
    however, implementing them is not required.
    """
    FEATURES = TIMELINE_FEATURES
    URL = 'https://twitter.com/...'

    _rate_limit: Optional[TwitterRateLimit] = None

    @classmethod
    def rate_limit(cls) -> TwitterRateLimit:
        if cls._rate_limit is None:
            cls._rate_limit = TwitterRateLimit(cls.__name__)
        return cls._rate_limit

    @staticmethod
    def get_base_variables(has_continuation: bool = False):
        raise NotImplementedError

    @classmethod
    def get_variables(cls, continuation: Optional[str] = None, count: int = 20, user_id: Optional[str] = None) -> str:
        variables = cls.get_base_variables(continuation is not None)
        if continuation is not None:
            variables['cursor'] = continuation
        variables['count'] = count
        if user_id is not None:
            variables['userId'] = user_id
        variables_text = json.dumps(variables)
        return variables_text

    @classmethod
    def prepare_for(cls, host, cookies, variables: str) -> RequestDetails:
        url = replace_url_host(cls.URL, host)

        params = {}
        params['variables'] = variables
        params['features'] = cls.FEATURES

        headers = get_auth_headers(cookies)

        details = RequestDetails(url=url, params=params, headers=headers, rate_limit=cls.rate_limit())
        return details

    @classmethod
    @abc.abstractmethod
    def prepare(cls, *args, **kwargs) -> RequestDetails:
        """Prepare a RequestDetails object based on passed arguments"""


class UserIDEndpoint(TwitterEndpoint):
    URL = EndpointUrl.USER_BY_SCREEN_NAME
    FEATURES = USER_FEATURES

    @classmethod
    def prepare(cls, host: str, cookies, user_handle: str) -> RequestDetails:
        user_handle = user_handle.strip('@/')
        variables = {'screen_name': user_handle, 'withSafetyModeUserFields': True}
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)

    @staticmethod
    def get_user_id(data: dict) -> Optional[str]:
        user_id = find_one(data, '$.data.user.result.rest_id')
        return str(user_id) if user_id is not None else None


class TweetDetailEndpoint(TwitterEndpoint):
    URL = EndpointUrl.TWEET_DETAIL
    FEATURES = TWEET_DETAIL_FEATURES

    @classmethod
    def prepare(cls, host: str, cookies, tweet_id: str, continuation: Optional[str] = None) -> RequestDetails:
        variables = {'focalTweetId': tweet_id, 'with_rux_injections': False, 'includePromotedContent': True,
                     'withCommunity': True, 'withQuickPromoteEligibilityTweetFields': True, 'withBirdwatchNotes': True,
                     'withVoice': True, 'withV2Timeline': True}
        if continuation:
            variables.update({'referrer': 'tweet', 'cursor': continuation})
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)


class AudioSpaceEndpoint(TwitterEndpoint):
    URL = EndpointUrl.AUDIOSPACE_BY_ID
    FEATURES = SPACE_FEATURES

    @classmethod
    def prepare(cls, host, cookies, space_id: str) -> RequestDetails:
        variables = {'id': space_id, 'isMetatagsQuery': False, 'withReplays': True, 'withListeners': True}
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)


class LiveStreamEndpoint(TwitterEndpoint):
    """Endpoint for retrieving HLS playlist for AudioSpace"""
    url = 'https://twitter.com/i/api/1.1/live_video_stream/status/{}'

    @classmethod
    def prepare(cls, host: str, cookies, media_key: str) -> RequestDetails:
        url = cls.url.format(media_key)
        url = replace_url_host(url, host)
        params = {'client': 'web', 'use_syndication_guest_id': 'false', 'cookie_set_host': get_netloc(host)}
        headers = get_auth_headers(cookies)

        details = RequestDetails(url=url, params=params, headers=headers, rate_limit=cls.rate_limit())
        return details


class LatestTimelineEndpoint(TwitterEndpoint):
    URL = EndpointUrl.LATEST_TIMELINE
    FEATURES = TIMELINE_FEATURES

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        if not is_continuation:
            request_context = 'launch'
        else:
            request_context = 'ptr'
        variables = {'includePromotedContent': True, 'latestControlAvailable': True, 'requestContext': request_context}
        return variables

    @classmethod
    def prepare(cls, host: str, cookies, continuation: Optional[str] = None, count: int = 20) -> RequestDetails:
        variables = cls.get_variables(continuation=continuation, count=count)
        return cls.prepare_for(host, cookies, variables)


class TimelineEndpoint(LatestTimelineEndpoint):
    URL = EndpointUrl.TIMELINE
    FEATURES = TIMELINE_FEATURES

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {"includePromotedContent": True, "latestControlAvailable": True, "withCommunity": True}


class UserTweetsEndpoint(TwitterEndpoint):
    URL = EndpointUrl.USER_TWEETS
    FEATURES = USER_TWEETS_FEATURES

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {'includePromotedContent': True, 'withQuickPromoteEligibilityTweetFields': True, 'withVoice': True, 'withV2Timeline': True}

    @classmethod
    def prepare(cls, host: str, cookies, user_id: str, continuation: Optional[str] = None, count: int = 20) -> RequestDetails:
        variables = cls.get_variables(user_id=user_id, count=count, continuation=continuation)
        return cls.prepare_for(host, cookies, variables)


class UserTweetsRepliesEndpoint(UserTweetsEndpoint):
    URL = EndpointUrl.USER_TWEETS_AND_REPLIES
    FEATURES = USER_TWEETS_FEATURES

    @staticmethod
    def get_base_variables(is_continuation: bool = False):
        return {'includePromotedContent': True, 'withCommunity': True, 'withVoice': True, 'withV2Timeline': True}


class SearchQueryType(str, enum.Enum):
    LATEST = 'Latest'
    TOP = 'Top'
    MEDIA = 'Media'


class SearchTimelineEndpoint(TwitterEndpoint):
    URL = EndpointUrl.SEARCH_TIMELINE
    FEATURES = SEARCH_FEATURES

    @classmethod
    def prepare(cls, host: str, cookies, raw_query: str, query_type: SearchQueryType, continuation: Optional[str] = None, count: int = 20) -> RequestDetails:
        variables = {'rawQuery': raw_query, 'count': count, 'querySource': 'typed_query', 'product': query_type.value}
        if continuation:
            variables.update({'cursor': continuation})
        variables_text = json.dumps(variables)
        return cls.prepare_for(host, cookies, variables_text)


def get_rate_limit_delay(headers: Union[Dict[str, str], CIMultiDictProxy[str]], logger: Optional[logging.Logger] = None) -> int:
    logger = logger or logging.getLogger().getChild('twitter_endpoints')
    retry_after = get_retry_after(headers)
    if retry_after is not None:
        return retry_after
    try:
        limit_total = int(headers.get('x-rate-limit-limit', -1))
        limit_remaining = int(headers.get('x-rate-limit-remaining', -1))
        reset_at = int(headers.get('x-rate-limit-reset', -1))
    except ValueError as e:
        logger.debug(f'error parsing limit headers: "{headers}"')
        return 0
    now = int(datetime.datetime.now().timestamp())
    reset_after = max(0, reset_at - now)
    logger.debug(f'rate limit {limit_remaining}/{limit_total}, resets after {reset_after} (at {reset_at})')
    if limit_remaining <= 1:
        return reset_after + 1
    return 0



import pytest

from avtdl.core.db import calculate_offset

testcases = {
    'first page': (1, 10, 100, (10, 90)),
    'second page': (2, 10, 100, (10, 80)),
    'last full page': (10, 10, 100, (10, 0)),
    'last page with less rows': (10, 10, 95, (5, 0)),
    'middle page': (3, 10, 50, (10, 20)),
    'no rows (first page)': (1, 10, 0, (0, 0)),
    'less than per_page': (1, 10, 5, (5, 0)),
    'exactly per_page': (1, 5, 5, (5, 0)),
}


@pytest.mark.parametrize('page, per_page, total_rows, expected', testcases.values(), ids=testcases.keys())
def test_offset(page, per_page, total_rows, expected):
    limit, offset = calculate_offset(page, per_page, total_rows)
    assert (limit, offset) == expected



import asyncio
import math
from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, List, Optional, Sequence

from pydantic import FilePath, NonNegativeFloat, field_validator

from avtdl.core.actors import Action, ActionEntity, ActorConfig, Monitor, MonitorEntity
from avtdl.core.db import RecordDB
from avtdl.core.interfaces import Record
from avtdl.core.plugins import Plugins
from avtdl.core.runtime import RuntimeContext, TaskStatus


class QuitMode(str, Enum):
    single = 'single'
    gather = 'gather'


@Plugins.register('utils.quit', Plugins.kind.ACTOR_CONFIG)
class QuitActionConfig(ActorConfig):
    pass


@Plugins.register('utils.quit', Plugins.kind.ACTOR_ENTITY)
class QuitActionEntity(ActionEntity):
    local_counter: int = -1
    """decrement counter on every received record, quit when it reaches zero"""
    global_counter: int = -1
    """decrement counter on every received record, quit when all global counters reaches zero"""


@Plugins.register('utils.quit', Plugins.kind.ACTOR)
class QuitAction(Action):
    """
    Used for testing purposes

    Triggers avtdl shutdown when specified number of records is received.
    """

    def __init__(self, conf: QuitActionConfig, entities: Sequence[QuitActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    def handle(self, entity: QuitActionEntity, record: Record):
        if entity.local_counter > 0:
            entity.local_counter -= 1
            self.logger.info(f'[{entity.name}] decrementing local counter, {entity.local_counter} left')
        if entity.local_counter == 0:
            self.quit(entity)

        if entity.global_counter > 0:
            entity.global_counter -= 1
            self.logger.info(
                f'[{entity.name}] decrementing global counter, {entity.global_counter} left for this entity')
        self.check_global_counter(entity)

    def check_global_counter(self, entity: QuitActionEntity):
        counters = []
        for entity in self.entities.values():  # type: ignore
            if entity.global_counter >= 0:
                counters.append(entity.global_counter)
        if all(c == 0 for c in counters):
            self.quit(entity)

    def quit(self, entity: QuitActionEntity):
        msg = f'[{entity.name}] interrupting program execution'
        self.logger.info(msg)
        raise KeyboardInterrupt(msg)


Plugins.register('utils.producer', Plugins.kind.ACTOR_CONFIG)(ActorConfig)

Plugins.register('utils.producer', Plugins.kind.ACTOR_ENTITY)(MonitorEntity)


@Plugins.register('utils.producer', Plugins.kind.ACTOR)
class Producer(Monitor):
    """
    Used for testing purposes

    Produces records programmatically.
    """

    def produce(self, entity_name: str, record: Record):
        """Programmatically emit given record on behalf of own entity with given name"""
        entity = self.entities[entity_name]
        self.on_record(entity, record)


Plugins.register('utils.consumer', Plugins.kind.ACTOR_CONFIG)(ActorConfig)

Plugins.register('utils.consumer', Plugins.kind.ACTOR_ENTITY)(ActionEntity)


@Plugins.register('utils.consumer', Plugins.kind.ACTOR)
class Consumer(Action):
    """
    Used for testing purposes

    Keeps history of received records, allow registering handlers
    for records received by specific entity.
    """

    def __init__(self, conf: ActorConfig, entities: Sequence[ActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.history: Dict[str, List[Record]] = {entity.name: [] for entity in entities}
        self.callbacks: Dict[Optional[str], List[Callable]] = defaultdict(list)

    def handle(self, entity: ActionEntity, record: Record):
        if entity.name in self.history:
            self.history[entity.name].append(record)
        for callback in self.callbacks[None]:
            callback(entity, record)
        if entity.name in self.callbacks:
            for callback in self.callbacks[entity.name]:
                callback(entity, record)

    def register_callback(self, callback: Callable[[ActionEntity, Record], None], entity_name: Optional[str] = None):
        self.callbacks[entity_name].append(callback)


Plugins.register('utils.replay', Plugins.kind.ACTOR_CONFIG)(ActorConfig)


@Plugins.register('utils.replay', Plugins.kind.ACTOR_ENTITY)
class ReplayEntity(MonitorEntity):
    db_path: FilePath
    """path to the sqlite database file storing records"""
    entity_name: Optional[str] = None
    """if specified, only records belonging to entity with this name are replayed"""
    emit_limit: Optional[int] = 10
    """how many records should be replayed. Use -1 to disable limit"""
    emit_interval: NonNegativeFloat = 0.01
    """delay between two consequentially produced records, in seconds"""
    reverse: bool = True
    """replace records from newest to oldest"""

    @field_validator('emit_limit')
    @classmethod
    def negative_is_none(cls, emit_limit):
        if emit_limit < 0:
            return None
        return emit_limit


@Plugins.register('utils.replay', Plugins.kind.ACTOR)
class Replay(Monitor):
    """
    Used for testing purposes

    Load records from db, emit them one by one
    """

    def __init__(self, conf: ActorConfig, entities: Sequence[ReplayEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.databases = {}
        for entity in entities:
            db = RecordDB(entity.db_path, logger=self.logger.getChild('name'))
            self.databases[entity.name] = db

    async def replay_task(self, entity: ReplayEntity, info: TaskStatus):
        db = self.databases.get(entity.name)
        if db is None:
            self.logger.exception(f'no database is opened for entity {entity.name}')
            return

        per_page = 20
        total_records = db.page_count(entity.entity_name, per_page) * per_page
        if entity.emit_limit == 0:
            return
        elif entity.emit_limit is None:
            to_emit = total_records
        else:
            to_emit = min(entity.emit_limit, total_records)
        pages_to_emit = max(1, math.ceil(to_emit / per_page))
        emitted = 0

        self.logger.debug(f'[{entity.name}] {to_emit} records to emit with {entity.emit_interval} interval')
        for page in range(pages_to_emit):
            records = db.load_page(entity.entity_name, page, per_page, entity.reverse)
            self.logger.debug(f'[{entity.name}] page {page} got {len(records)} records')
            for record in records:
                if emitted >= to_emit:
                    break
                emitted += 1
                self.on_record(entity, record)
                info.set_status(f'{emitted}/{to_emit} records replayed', record)
                await asyncio.sleep(entity.emit_interval)
        self.logger.debug(f'[{entity.name}] done')
        info.clear()

    async def run(self):
        for entity in self.entities.values():
            info = TaskStatus(self.conf.name, entity.name)
            task = self.replay_task(entity, info)
            self.ctx.controller.create_task(task, name=f'{self.conf.name}:{entity.name}', _info=info)
        await super().run()



import asyncio
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Sequence

from pydantic import BaseModel, FilePath, NonNegativeFloat

from avtdl.core.actors import Action, ActionEntity, ActorConfig
from avtdl.core.config import SettingsSection
from avtdl.core.formatters import sanitize_filename
from avtdl.core.interfaces import Record
from avtdl.core.request import HttpClient, SessionStorage
from avtdl.core.runtime import RuntimeContext, TaskStatus
from avtdl.core.state import StateSerializer
from avtdl.core.utils import ListRootModel, check_dir, with_prefix


class HttpActionConfig(ActorConfig):
    pass


class HttpActionEntity(ActionEntity):
    cookies_file: Optional[FilePath] = None
    """path to a text file containing cookies in Netscape format"""
    headers: Optional[Dict[str, str]] = {}
    """custom HTTP headers as pairs "key": value". "Set-Cookie" header will be ignored, use `cookies_file` option instead"""


class HttpAction(Action, ABC):

    def __init__(self, conf: HttpActionConfig, entities: Sequence[HttpActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.conf: HttpActionConfig
        self.entities: Mapping[str, HttpActionEntity]  # type: ignore
        self.sessions = SessionStorage(self.logger)
        settings: Optional[SettingsSection] = ctx.get_extra('settings')
        if settings is None:
            raise RuntimeError(f'runtime context is missing Settings instance. This is a bug, please report it')
        self.settings = settings

    async def run(self) -> None:
        name = f'ensure_closed for {self.logger.name} ({self!r})'
        _ = self.controller.create_task(self.sessions.ensure_closed(), name=name)
        await super().run()

    def get_client(self, entity: HttpActionEntity) -> HttpClient:
        """provide  HttpClient instance for entity task to make network requests"""
        session = self.sessions.get_session(entity.cookies_file, entity.headers)
        logger = with_prefix(self.logger, f'[{entity.name}] ')
        client = HttpClient(logger, session)
        return client


class QueueActionConfig(HttpActionConfig):
    consumption_delay: NonNegativeFloat = 1
    """delay before entity starts processing next record after finishing previous, in seconds"""


class QueueActionEntity(HttpActionEntity):
    restartable: bool = True
    """Attempt to store unprocessed records on disk at shutdown and process them on the next startup"""


class QueueAction(HttpAction):

    def __init__(self, conf: QueueActionConfig, entities: Sequence[QueueActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

        self.conf: QueueActionConfig
        self.entities: Mapping[str, QueueActionEntity]  # type: ignore
        self.queues: Dict[str, asyncio.Queue] = defaultdict(lambda: asyncio.Queue())
        self.info: Dict[str, TaskStatus] = {entity.name: TaskStatus(self.conf.name, entity.name) for entity in entities}

    def persistence_file(self, entity_name: str) -> Path:
        filename = sanitize_filename(entity_name)
        file = f'{self.conf.name}/{filename}.dat'
        return Path(file)

    def load_queue(self, entity: QueueActionEntity):
        if not entity.restartable:
            return
        persistence_path = self.settings.state_directory / self.persistence_file(entity.name)
        serialized = StateSerializer.restore(QueueSerialized, persistence_path)
        if serialized is None:
            return
        queue = serialized.to_queue()
        self.queues[entity.name] = queue
        self.logger.debug(f'[{entity.name}] restored {queue.qsize()} unprocessed records from the previous run')

    def dump_queue(self, entity: QueueActionEntity):
        if not entity.restartable:
            return
        queue = self.queues[entity.name]
        if queue.qsize() == 0:
            return
        serialized = QueueSerialized.from_queue(queue)
        persistence_path = self.settings.state_directory / self.persistence_file(entity.name)
        ok = StateSerializer.dump(serialized, persistence_path)
        if ok:
            self.logger.debug(f'[{entity.name}] stored {len(serialized)} unprocessed records until the next run')

    def handle(self, entity: ActionEntity, record: Record):
        try:
            queue = self.queues[entity.name]
            queue.put_nowait(record)
        except (asyncio.QueueFull, KeyError) as e:
            self.logger.exception(
                f'[{entity.name}] failed to add url, {type(e)}: {e}. This is a bug, please report it.')
        else:
            self.logger.debug(f'[{entity.name}] added new record to the queue, current queue size is {queue.qsize()}')

    async def run_for(self, entity: QueueActionEntity):
        logger = with_prefix(self.logger, f'[{entity.name}] ')
        client = self.get_client(entity)
        self.load_queue(entity)
        queue = self.queues[entity.name]
        try:
            while True:
                record = await queue.get()
                self.logger.debug(f'(queued: {queue.qsize()}) processing record {record!r}')
                await self.handle_single_record(logger, client, entity, record)
                await asyncio.sleep(self.conf.consumption_delay)
                self.update_info(entity, record)
        except Exception:
            logger.exception(f'unexpected error in background task, terminating')
        except (KeyboardInterrupt, asyncio.CancelledError):
            self.dump_queue(entity)
            raise

    def update_info(self, entity: QueueActionEntity, record: Record):
        info = self.info.get(entity.name)
        if info is None:
            return
        queue = self.queues[entity.name]
        size = queue.qsize()
        if size:
            info.set_status(f'current queue size is {size}', record)
        else:
            info.clear()

    async def run(self) -> None:
        for entity in self.entities.values():
            name = f'{self.conf.name}:{entity.name}'
            info = self.info.get(entity.name)
            _ = self.controller.create_task(self.run_for(entity), name=name, _info=info)
        await super().run()

    @abstractmethod
    async def handle_single_record(self, logger: logging.Logger, client: HttpClient,
                                   entity: QueueActionEntity, record: Record) -> None:
        """Called for each record waiting in specific entity's queue"""


class QueueSerialized(ListRootModel):
    root: List[Record] = []

    def to_queue(self) -> asyncio.Queue:
        queue: asyncio.Queue = asyncio.Queue()
        for item in reversed(self.root):
            try:
                queue.put_nowait(item)
            except asyncio.QueueFull:
                break
        return queue

    @classmethod
    def from_queue(cls, queue: asyncio.Queue):
        records = []
        while True:
            try:
                records.append(queue.get_nowait())
            except asyncio.QueueEmpty:
                break
        return cls(root=records)


class TaskActionConfig(HttpActionConfig):
    consumption_delay: NonNegativeFloat = 1
    """delay between start of processing of multiple records received at the same time, in seconds. Used to even out short bursts of activity"""


class TaskActionEntity(HttpActionEntity):
    restartable: bool = True
    """Attempt to store unprocessed records on disk at shutdown and restart tasks to process them on the next startup"""


class TaskAction(HttpAction):

    def __init__(self, conf: TaskActionConfig, entities: Sequence[TaskActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.conf: TaskActionConfig
        self.entities: Mapping[str, TaskActionEntity]  # type: ignore
        self.tasks: Dict[str, asyncio.Task] = {}
        self.start_token = asyncio.Lock()

    def handle(self, entity: TaskActionEntity, record: Record):
        logger = with_prefix(self.logger, f'[{entity.name}]')
        task_name = self.task_name_for(entity, record)
        if task_name in self.tasks:
            logger.debug(f'task for record {record!r} is already running')
            return
        info = TaskStatus(self.conf.name, entity.name, record=record)
        client = self.get_client(entity)
        task = self.controller.create_task(self._handle_record_task(logger, client, entity, record, info), name=task_name, _info=info)
        task.add_done_callback(lambda _: self.tasks.pop(task_name))
        self.tasks[task_name] = task

    def task_name_for(self, entity: TaskActionEntity, record: Record) -> str:
        record_id = record.get_uid()
        record_hash = record.hash()
        task_name = f'{self.conf.name}:{entity.name} {record_id} {record_hash}'
        return task_name

    async def _handle_record_task(self, logger: logging.Logger, client: HttpClient,
                                  entity: TaskActionEntity, record: Record, info: TaskStatus) -> None:
        async with self.start_token:
            # ideally delay should be applied after the task creation, but it means adding yet another create_task()
            await asyncio.sleep(self.conf.consumption_delay)
        try:
            await self.handle_record_task(logger, client, entity, record, info)
        except Exception:
            logger.exception(f'unexpected exception while processing record {record!r}')
        except (KeyboardInterrupt, asyncio.CancelledError):
            self.dump_task(entity, record)
            raise

    @abstractmethod
    async def handle_record_task(self, logger: logging.Logger, client: HttpClient,
                                 entity: TaskActionEntity, record: Record, info: TaskStatus) -> None:
        """Scheduled as task for each record to be processed"""

    async def run(self) -> None:
        self.load_tasks()
        await super().run()

    def persistence_directory(self) -> Path:
        return self.settings.state_directory / self.conf.name

    def persistence_path_for(self, entity: TaskActionEntity, record: Record) -> Path:
        name = sanitize_filename(self.task_name_for(entity, record))
        return self.persistence_directory() / f'{name}.dat'

    def load_tasks(self) -> None:
        if not check_dir(self.persistence_directory(), False):
            return
        try:
            files = list(self.persistence_directory().iterdir())
        except OSError as e:
            self.logger.warning(f'failed to list files in "{self.persistence_directory()}": {e}')
            return
        for file in files:
            task_to_restore = StateSerializer.restore(TaskSerialized, file)
            if task_to_restore is None:
                continue
            restored_entity, record = task_to_restore.entity, task_to_restore.record
            self.logger.debug(f'[{restored_entity.name}] found persistence data for record {record!r} in {file}')
            try:
                file.unlink(missing_ok=True)
                self.logger.debug(f'[{restored_entity.name}] persistence data in {file} successfully deleted')
            except OSError as e:
                self.logger.warning(f'failed to delete task persistence data in {file}: {e}')
            live_entity: Optional[TaskActionEntity] = self.entities.get(restored_entity.name)
            if live_entity is None:
                self.logger.debug(f'entity "{restored_entity.name}" is not present in config after restart, task for record "{record!r}" will not be restarted')
                continue
            if not live_entity.restartable:
                self.logger.debug(f'[{live_entity.name}] restarting disabled for this entity, task for record "{record!r}" will not be restarted')
                continue
            if restored_entity != live_entity:
                self.logger.debug(f'[{live_entity.name}] re-running task for entity that has changed: "{restored_entity}" to "{live_entity}"')
            self.logger.info(f'[{live_entity.name}] re-running task for record "{record!r}"')
            self.handle_record(live_entity, record)

    def dump_task(self, entity: TaskActionEntity, record: Record) -> None:
        if not entity.restartable:
            return
        serialized = TaskSerialized(entity=entity, record=record)
        path = self.persistence_path_for(entity, record)
        ok = StateSerializer.dump(serialized, path)
        if ok:
            self.logger.info(f'[{entity.name}] task for record "{record!r}" is stored to be re-run after restart')

class TaskSerialized(BaseModel):
    entity: TaskActionEntity
    record: Record


#!/usr/bin/env python3
import datetime
from enum import Enum
from typing import Callable, Dict, List, Sequence

from avtdl.core.config import Plugins
from avtdl.core.interfaces import Event, Record, TextRecord
from avtdl.core.monitors import BaseFeedMonitor, BaseFeedMonitorConfig, BaseFeedMonitorEntity
from avtdl.core.request import DataResponse, HttpClient, HttpResponse
from avtdl.core.utils import sha1, utcnow

Plugins.register('get_url', Plugins.kind.ASSOCIATED_RECORD)(TextRecord)


@Plugins.register('get_url', Plugins.kind.ASSOCIATED_RECORD)
class UrlStatusRecord(TextRecord):
    url: str
    """url of the monitored page"""
    text: str
    """content of the web page. Might be replaced with empty string to save space"""
    text_hash: str
    """hash of the web page content. On network error hash of the last successfully fetched page is used"""
    accessed_at: datetime.datetime
    """time the url was fetched at"""
    ok: bool
    """set to True if HTTP request completed successfully"""
    status: int
    """HTTP response status. Special value 0 means network connection failed"""
    reason: str
    """short text explaining status code provided by the server or error if network connection failed"""
    headers: Dict[str, str] = {}
    """response headers"""

    def get_uid(self) -> str:
        return self.url

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f'UrlStatusRecord(url="{self.url}", accessed_at="{self.accessed_at}", status="{self.status}", reason="{self.reason}", text_hash="{self.text_hash}")'


@Plugins.register('get_url', Plugins.kind.ACTOR_CONFIG)
class UrlMonitorConfig(BaseFeedMonitorConfig):
    pass


@Plugins.register('get_url', Plugins.kind.ACTOR_ENTITY)
class UrlMonitorEntity(BaseFeedMonitorEntity):
    url: str
    """url to monitor"""
    store_text: bool = True
    """store page content in the "text" field of the record"""
    report_server_error: bool = False
    """emit an event if error status has been received in HTTP response, as well as when the url started responding with 2xx again"""
    report_network_error: bool = False
    """emit an event if a network error has occurred and when the request succeeded again after the network outage"""


@Plugins.register('get_url', Plugins.kind.ACTOR)
class UrlMonitor(BaseFeedMonitor):
    """
    Monitor web page state

    Download contents of web page at `url` and emit it as an `UrlStatusRecord`
    if it has changed since the last update. Intended to be used with simple
    text endpoints.
    """

    async def get_records(self, entity: UrlMonitorEntity, client: HttpClient) -> Sequence[UrlStatusRecord]:
        response = await self.request_raw(entity.url, entity, client)
        record = UrlStatusRecord(
            url=entity.url,
            text=(response.text or '') if entity.store_text else '',
            text_hash=sha1(response.text) if isinstance(response, DataResponse) else '',
            accessed_at=utcnow(),
            ok=response.ok,
            status=response.status,
            reason=response.reason if isinstance(response, HttpResponse) else str(response.e),
            headers={k: v for k, v in response.headers.items()} if response.headers is not None else {}
        )
        return [record]

    def filter_new_records(self, records: Sequence[UrlStatusRecord], entity: UrlMonitorEntity) -> Sequence[Record]:
        new_records: List[Record] = []
        for record in records:
            new_records.extend(self.filter_single_record(record, entity))
        return new_records

    def filter_single_record(self, record: UrlStatusRecord, entity: UrlMonitorEntity) -> Sequence[Record]:
        new_records: List[Record] = []
        stored_record = self.load_record(record, entity)
        if stored_record is None:
            self.logger.debug(
                f'[{entity.name}] fetched record is new: "{record.get_uid()}" (hash: {record.hash()[:5]})')
            return [record]
        if not isinstance(stored_record, UrlStatusRecord):
            self.logger.warning(f'[{entity.name} loaded record has unexpected type: {stored_record!r}')
            return [record]
        status = f'{record.status} {record.reason}'
        stored_status = f'{stored_record.status} {stored_record.reason}'
        if record.ok:
            if record.text_hash != stored_record.text_hash:
                self.store_records([record], entity)
                new_records.append(record)
                self.logger.debug(
                    f'[{entity.name}] storing new version of record "{record.get_uid()}" (hash: {record.hash()[:5]})')
            if not stored_record.ok:
                if stored_record.status == 0:
                    if entity.report_network_error:
                        event = Event(event_type='status',
                                      text=f'{record.url} is back online with {status} after network outage')
                        new_records.append(event)
                else:
                    if entity.report_server_error:
                        event = Event(event_type='status',
                                      text=f'{record.url} responded with {status} after {stored_status}')
                        new_records.append(event)
        else:
            record.text_hash = stored_record.text_hash
            if record.status == 0 and stored_record.status != 0:
                if entity.report_network_error:
                    event = Event(event_type='status',
                                  text=f'got network error requesting {record.url}: {record.reason}')
                    new_records.append(event)
            elif record.status != 0 and stored_record.status == 0:
                if entity.report_network_error:
                    event = Event(event_type='status',
                                  text=f'{record.url} is back online with {status} after network outage')
                    new_records.append(event)
            elif record.status != stored_record.status:
                if entity.report_server_error:
                    event = Event(event_type='status',
                                  text=f'{record.url} responded with {status} after {stored_status}')
                    new_records.append(event)
        return new_records

    def filter_single_record_a(self, record: UrlStatusRecord, entity: UrlMonitorEntity) -> Sequence[Record]:
        def store_new(record: UrlStatusRecord, stored_record: UrlStatusRecord) -> Sequence[Record]:
            if record.text_hash != stored_record.text_hash:
                self.store_records([record], entity)
                return [record]
            return []

        def copy_hash(record: UrlStatusRecord, stored_record: UrlStatusRecord) -> Sequence:
            record.text_hash = stored_record.text_hash
            return []

        def make_event(condition: bool, msg: str) -> Callable[[UrlStatusRecord, UrlStatusRecord], Sequence[Event]]:
            def event_factory(*_) -> Sequence[Event]:
                return [Event(event_type='status', text=msg)] if condition else []

            return event_factory

        class UrlStatus(str, Enum):
            ok = 'OK'
            bad = 'BAD'
            error = 'NETWORK ERROR'

        def record_status(record: UrlStatusRecord) -> UrlStatus:
            if record.ok:
                return UrlStatus.ok
            elif record.status == 0:
                return UrlStatus.error
            else:
                return UrlStatus.bad

        stored_record = self.load_record(record, entity)
        if stored_record is None:
            self.logger.debug(
                f'[{entity.name}] fetched record is new: "{record.get_uid()}" (hash: {record.hash()[:5]})')
            return [record]
        if not isinstance(stored_record, UrlStatusRecord):
            self.logger.warning(f'[{entity.name} loaded record has unexpected type: {stored_record!r}')
            return [record]
        status = f'{record.status} {record.reason}'
        stored_status = f'{stored_record.status} {stored_record.reason}'
        # mapping of record state to stored_record state pairs to actions to be taken
        actions: Dict[
            UrlStatus, Dict[UrlStatus, List[Callable[[UrlStatusRecord, UrlStatusRecord], Sequence[Record]]]]] = {
            UrlStatus.ok: {
                UrlStatus.ok: [store_new],
                UrlStatus.bad: [
                    store_new,
                    make_event(entity.report_server_error,
                               f'{record.url} responded with {status} after {stored_status}')
                ],
                UrlStatus.error: [
                    store_new,
                    make_event(entity.report_network_error,
                               f'{record.url} is back online with {status} after network outage')
                ],
            },
            UrlStatus.bad: {
                UrlStatus.ok: [
                    copy_hash,
                    make_event(entity.report_server_error,
                               f'{record.url} responded with {status} after {stored_status}')
                ],
                UrlStatus.bad: [
                    copy_hash,
                    make_event(entity.report_server_error and record.status != stored_record.status,
                               f'{record.url} responded with {status} after {stored_status}')
                ],
                UrlStatus.error: [
                    copy_hash,
                    make_event(entity.report_network_error,
                               f'{record.url} is back online with {status} after network outage'),
                    make_event(entity.report_server_error,
                               f'{record.url} responded with {status} after network outage')
                ],
            },
            UrlStatus.error: {
                UrlStatus.ok: [
                    copy_hash,
                    make_event(entity.report_network_error,
                               f'got network error requesting {record.url}: {record.reason}')
                ],
                UrlStatus.bad: [
                    copy_hash,
                    make_event(entity.report_network_error,
                               f'got network error requesting {record.url}: {record.reason}')
                ],
                UrlStatus.error: [copy_hash],
            },
        }
        action_sequence = actions[record_status(record)][record_status(stored_record)]
        new_records: List[Record] = []
        for action in action_sequence:
            new_records.extend(action(record, stored_record))
        return new_records

    def filter_single_record_b(self, record: UrlStatusRecord, entity: UrlMonitorEntity) -> Sequence[Record]:
        new_records: List[Record] = []
        stored_record = self.load_record(record, entity)
        if stored_record is None:
            self.logger.debug(
                f'[{entity.name}] fetched record is new: "{record.get_uid()}" (hash: {record.hash()[:5]})')
            return [record]
        if not isinstance(stored_record, UrlStatusRecord):
            self.logger.warning(f'[{entity.name} loaded record has unexpected type: {stored_record!r}')
            return [record]

        class UrlStatus(str, Enum):
            ok = 'OK'
            bad = 'BAD'
            error = 'NETWORK ERROR'

        def record_status(record: UrlStatusRecord) -> UrlStatus:
            if record.ok:
                return UrlStatus.ok
            elif record.status == 0:
                return UrlStatus.error
            else:
                return UrlStatus.bad

        def add_event(condition: bool, msg: str):
            if condition:
                new_records.append(Event(event_type='status', text=msg))

        status = f'{record.status} {record.reason}'
        stored_status = f'{stored_record.status} {stored_record.reason}'

        if record_status(record) == UrlStatus.ok:
            if record.text_hash != stored_record.text_hash:
                self.store_records([record], entity)
                new_records.append(record)
                self.logger.debug(
                    f'[{entity.name}] storing new version of record "{record.get_uid()}" (hash: {record.hash()[:5]})')

            if record_status(stored_record) == UrlStatus.ok:
                pass
            elif record_status(stored_record) == UrlStatus.bad:
                add_event(entity.report_network_error,
                          f'{record.url} is back online with {status} after network outage')
            elif record_status(stored_record) == UrlStatus.error:
                add_event(entity.report_server_error, f'{record.url} responded with {status} after {stored_status}')

        elif record_status(record) == UrlStatus.bad:
            record.text_hash = stored_record.text_hash

            if record_status(stored_record) == UrlStatus.ok:
                add_event(entity.report_server_error, f'{record.url} responded with {status} after {stored_status}')
            elif record_status(stored_record) == UrlStatus.bad:
                add_event(entity.report_server_error and record.status != stored_record.status,
                          f'{record.url} responded with {status} after {stored_status}')
            elif record_status(stored_record) == UrlStatus.error:
                add_event(entity.report_network_error,
                          f'{record.url} is back online with {status} after network outage'),
                add_event(entity.report_server_error,
                          f'{record.url} responded with {status} after network outage')

        elif record_status(record) == UrlStatus.error:
            record.text_hash = stored_record.text_hash

            if record_status(stored_record) == UrlStatus.ok:
                add_event(entity.report_network_error, f'got network error requesting {record.url}: {record.reason}')
            elif record_status(stored_record) == UrlStatus.bad:
                add_event(entity.report_network_error, f'got network error requesting {record.url}: {record.reason}')
            elif record_status(stored_record) == UrlStatus.error:
                pass

        return new_records



import datetime
from typing import Dict, Optional, Sequence

from pydantic import field_serializer, field_validator

from avtdl.core.actors import Filter, FilterEntity
from avtdl.core.formatters import Fmt
from avtdl.core.interfaces import Event, OpaqueRecord, Record, TextRecord
from avtdl.core.plugins import Plugins
from avtdl.core.runtime import RuntimeContext
from avtdl.core.utils import Timezone
from avtdl.plugins.filters.filters import EmptyFilterConfig

Plugins.register('filter.format', Plugins.kind.ASSOCIATED_RECORD)(TextRecord)
Plugins.register('filter.format', Plugins.kind.ACTOR_CONFIG)(EmptyFilterConfig)


class BaseFormatFilterEntity(FilterEntity):
    missing: Optional[str] = None
    """if specified, will be used to fill template placeholders that do not have corresponding fields in current record"""
    timezone: Optional[datetime.tzinfo] = None
    """takes timezone name from <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones> (or local time if omitted), converts record fields containing date and time to this timezone"""

    @field_validator('timezone', mode='plain')
    @classmethod
    def check_timezone(cls, timezone: Optional[str]) -> Optional[datetime.tzinfo]:
        if timezone is None:
            return None
        if not isinstance(timezone, str):
            raise ValueError('Input should be a valid string')
        return Timezone.get_tz(timezone)

    @field_serializer('timezone')
    @classmethod
    def serialize_timezone(cls, timezone: Optional[datetime.tzinfo]) -> Optional[str]:
        return Timezone.get_name(timezone)


@Plugins.register('filter.format', Plugins.kind.ACTOR_ENTITY)
class FormatFilterEntity(BaseFormatFilterEntity):
    template: str
    """template string with placeholders that will be filled with corresponding values from current record"""


@Plugins.register('filter.format', Plugins.kind.ACTOR)
class FormatFilter(Filter):
    """
    Format record as text

    Takes a record and produces a new `TextRecord` by taking `template` string
    and replacing "{placeholder}" with value of `placeholder` field of the
    current record, where `placeholder` is any field the record might have.
    If one of the placeholders is not a field of a specific record, it will be
    replaced with a value defined in `missing` parameter if it is specified,
    otherwise it will be left intact.

    Because output record is essentially a text, timezone offset will be lost
    for all fields containing date and time value. Therefore, the `timezone`
    parameter is provided to allow formatting these fields in desired timezone.
    """

    def __init__(self, config: EmptyFilterConfig, entities: Sequence[FormatFilterEntity], ctx: RuntimeContext):
        super().__init__(config, entities, ctx)

    def match(self, entity: FormatFilterEntity, record: Record) -> TextRecord:
        record = record.as_timezone(entity.timezone)
        text = Fmt.format(entity.template, record, entity.missing, tz=entity.timezone)
        return TextRecord(text=text)


Plugins.register('filter.format.event', Plugins.kind.ASSOCIATED_RECORD)(Event)
Plugins.register('filter.format.event', Plugins.kind.ACTOR_CONFIG)(EmptyFilterConfig)


@Plugins.register('filter.format.event', Plugins.kind.ACTOR_ENTITY)
class FormatEventFilterEntity(BaseFormatFilterEntity):
    type_template: str
    """template string with placeholders used to format "event_type" field of the produced event with corresponding values from current record"""
    text_template: str
    """template string with placeholders used to format "text" field of the produced event with corresponding values from current record"""


@Plugins.register('filter.format.event', Plugins.kind.ACTOR)
class FormatEventFilter(Filter):
    """
    Generate Event from record text

    Takes a record and produces a new `Event` with the `event_type` and
    `text` values evaluated by filling placeholders in `type_template`
    and `text_template`.

    Just like with the `filter.format`, missing placeholders are filled with
    value from the `missing` parameter if it is specified, otherwise they are
    left unchanged.

    Specifying timezone offset for the output text is also
    supported, though original record might be retrieved with `filter.event.cause`.
    """

    def __init__(self, config: EmptyFilterConfig, entities: Sequence[FormatEventFilterEntity], ctx: RuntimeContext):
        super().__init__(config, entities, ctx)

    def match(self, entity: FormatEventFilterEntity, record: Record) -> Event:
        record = record.as_timezone(entity.timezone)
        event_type = Fmt.format(entity.type_template, record, entity.missing, tz=entity.timezone)
        text = Fmt.format(entity.text_template, record, entity.missing, tz=entity.timezone)
        return Event(event_type=event_type, text=text, record=record)


Plugins.register('filter.format.opaque', Plugins.kind.ASSOCIATED_RECORD)(OpaqueRecord)
Plugins.register('filter.format.opaque', Plugins.kind.ACTOR_CONFIG)(EmptyFilterConfig)


@Plugins.register('filter.format.opaque', Plugins.kind.ACTOR_ENTITY)
class FormatOpaqueFilterEntity(BaseFormatFilterEntity):
    fields_templates: Dict[str, str]
    """mapping of fields names and corresponding template string with placeholders"""


@Plugins.register('filter.format.opaque', Plugins.kind.ACTOR)
class FormatOpaqueFilter(Filter):
    """
    Generate record with custom fields

    Takes a record and produces a new `OpaqueRecord` with fields names
    and values defined in the `fields_templates` mapping
    """

    def __init__(self, config: EmptyFilterConfig, entities: Sequence[FormatOpaqueFilterEntity], ctx: RuntimeContext):
        super().__init__(config, entities, ctx)

    def match(self, entity: FormatOpaqueFilterEntity, record: Record) -> OpaqueRecord:
        record = record.as_timezone(entity.timezone)
        fields = {}
        for name, fmt in entity.fields_templates.items():
            fields[name] = Fmt.format(fmt, record, entity.missing, entity.timezone)
        return OpaqueRecord(**fields)



import asyncio
import logging
import os
import re
import shutil
from pathlib import Path
from typing import List, Mapping, Optional, Sequence

from pydantic import AnyUrl, Field, NonNegativeFloat, RootModel, ValidationError, field_validator, model_validator

from avtdl.core.actions import QueueAction, QueueActionConfig, QueueActionEntity
from avtdl.core.cache import FileCache, find_free_suffix, find_with_suffix
from avtdl.core.config import SettingsSection
from avtdl.core.download import RemoteFileInfo, download_file, has_same_content, remove_files
from avtdl.core.formatters import Fmt, sanitize_filename
from avtdl.core.interfaces import Record
from avtdl.core.plugins import Plugins
from avtdl.core.request import HttpClient
from avtdl.core.runtime import RuntimeContext
from avtdl.core.utils import check_dir, is_url, sha1


@Plugins.register('download', Plugins.kind.ACTOR_CONFIG)
class FileDownloadConfig(QueueActionConfig):
    max_concurrent_downloads: int = Field(default=1, ge=1)
    """limit for simultaneously active download tasks among all entities. Note that each entity will still process records sequentially regardless of this setting"""
    partial_file_suffix: str = '.part'
    """appended to a name of the file that is not yet completely downloaded"""


@Plugins.register('download', Plugins.kind.ACTOR_ENTITY)
class FileDownloadEntity(QueueActionEntity):
    url_field: str
    """field in the incoming record containing url of file to be downloaded"""
    path: Path
    """directory where downloaded file should be created. Supports templating with {...}"""
    filename: Optional[str] = None
    """name downloaded file should be stored under. If not provided will be inferred from HTTP headers or download url. Supports templating with {...} (additionally, "{source_name}" placeholder will be replaced with the inferred value)"""
    extension: Optional[str] = None
    """normally file extension will be inferred from HTTP headers. This option allows to overwrite it"""
    overwrite: bool = False
    """whether file should be overwritten in if it already exists. If set to false will cause suffix with a number be added to the newly downloaded file name"""
    rename_suffix: str = ' [{i}]'
    """when overwriting is disabled, this suffix is attached to the base filename with the "{i}" part replaced with a number. Must contain "{i}" exactly once"""

    @field_validator('extension')
    @classmethod
    def ensure_dot(cls, value: Optional[str]) -> Optional[str]:
        if value is None:
            return None
        if value.startswith('.'):
            return value
        return '.' + value

    @field_validator('rename_suffix')
    @classmethod
    def check_suffix(cls, value: str) -> str:
        found = len(re.findall(r'{i}', value))
        if found != 1:
            raise ValueError('rename_suffix must contain exactly one occurrence of "{i}", got ' + str(found))
        value = sanitize_filename(value)
        return value


@Plugins.register('download', Plugins.kind.ACTOR)
class FileDownload(QueueAction):
    """
    Download a file

    Take an url from a field of a processed record with a name specified in `url_field`
    and download it as a file to specified location.
    The field must be present in the record and  must contain a valid url with a scheme
    (such as "https") or a list of such urls.

    Primarily designed for downloading images attached to a post, or thumbnails.
    Does not support resuming interrupted downloads or detecting that this exact file is
    already stored at target location without downloading it again.

    File extension and name are inferred from HTTP headers and path part of the url,
    unless provided explicitly with `extension` and `filename` parameters.
    Since final file name is determined as a part of the download process, the file
    is initially stored under a temporary name (currently an SHA1 of the url) in the
    download directory, and then renamed to target filename.

    If a file with given name already exists, depending on an `overwrite` setting a new file will either
    overwrite it or get stored under different name, generated by combining
    the base name with a number added as part of `rename_suffix`.
    If, however, an exact copy of the new file is found among the files in target directory
    sharing the base name, the new file will be deleted, giving preference to the existing copy.
    """

    def __init__(self, conf: FileDownloadConfig, entities: Sequence[FileDownloadEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.conf: FileDownloadConfig
        self.entities: Mapping[str, FileDownloadEntity]
        self.concurrency_limit = asyncio.BoundedSemaphore(value=conf.max_concurrent_downloads)

    def _get_urls_list(self, entity: FileDownloadEntity, record: Record) -> Optional[List[str]]:
        field = getattr(record, entity.url_field, None)
        if field is None:
            msg = f'received a record that does not contain "{entity.url_field}" field. The record: {record!r}'
            self.logger.debug(msg)
            return None
        if isinstance(field, str):
            field = [field]
        try:
            urls = [str(url) for url in UrlList(field)]
            return urls
        except ValidationError:
            self.logger.debug(
                f'received record with the "{entity.url_field}" field that was not recognised as a valid url or a sequence of urls. Raw record: {record!r}')
            return None

    async def handle_single_record(self, logger: logging.Logger, client: HttpClient,
                                   entity: FileDownloadEntity, record: Record) -> None:
        urls = self._get_urls_list(entity, record)
        if urls is None:
            self.logger.debug(f'found no values in field "{entity.url_field}", skipping record')
            return
        for url in urls:
            self.logger.debug(f'processing url {url}')
            await self.handle_download(logger, client, entity, record, url)

    async def handle_download(self, logger, client: HttpClient, entity: FileDownloadEntity, record: Record, url: str):
        """
        Handle download-related stuff: generating filenames, moving files, error reporting

        - generate tempfile name from url hash
        - if it exists abort
        - perform the download into a temp file
        - generate resulting file name
        - if exists either rename or replace depending on settings
        """
        path = Fmt.format_path(entity.path, record, tz=entity.timezone)
        ok = check_dir(path)
        if not ok:
            logger.warning(f'check "{path}" is a valid and writeable directory')
            return

        temp_file = path / Path(sha1(url)).with_suffix(self.conf.partial_file_suffix)
        if temp_file.exists():
            logger.warning(
                f'aborting download of "{url}": temporary file "{temp_file}" already exists, meaning download is already in progress or download process has been interrupted abruptly')
            return

        logger.debug(f'downloading "{url}" to "{temp_file}"')
        info = await self.download(logger, client, url, temp_file)
        if info is None:
            return None

        if entity.filename is not None:
            extra = {'source_name': info.source_name}
            filename = Fmt.format(entity.filename, record, tz=entity.timezone, extra=extra)
        else:
            filename = info.source_name
        filename = sanitize_filename(filename)
        path = path.joinpath(filename)
        if entity.extension is not None:
            path = path.with_suffix(entity.extension)
        else:
            path = path.with_suffix(info.extension)

        try:
            path.exists()
        except OSError as e:
            logger.warning(f'failed to process record: {e}')
            return
        if path.exists() and not entity.overwrite:
            for p in path.parent.iterdir():
                if not p.stem.startswith(path.stem):
                    continue
                if has_same_content(temp_file, p):
                    self.logger.info(f'file "{temp_file}" is already stored as "{p}", deleting')
                    remove_files([temp_file])
                    return
            new_path = Path(path)  # making a copy
            i = 0
            while new_path.exists():
                i += 1
                suffix = entity.rename_suffix.replace('{i}', str(i))
                new_name = path.stem + suffix
                new_path = new_path.with_stem(new_name)
            path = new_path
        move_file(temp_file, path, logger)

    async def download(self, logger: logging.Logger, client: HttpClient,
                       url: str, output_file: Path) -> Optional[RemoteFileInfo]:
        """Perform the actual download"""
        return await download(self.concurrency_limit, logger, client, url, output_file)


async def download(semaphore: asyncio.BoundedSemaphore, logger: logging.Logger, client: HttpClient,
                   url: str, output_file: Path) -> Optional[RemoteFileInfo]:
    try:
        async with semaphore:
            logger.debug(
                f'acquired semaphore({semaphore._value}), downloading "{url}" to "{output_file}"')
            info = await download_file(url, output_file, client.session)
    except Exception as e:
        logger.exception(f'unexpected error when downloading "{url}" to "{output_file}": {e}')
        return None
    logger.debug(
        f'finished downloading "{url}" to "{output_file}", semaphore({semaphore._value}) released')
    return info


class UrlList(RootModel):
    root: Sequence[AnyUrl]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]


def move_file(source: Path, target: Path, logger: logging.Logger) -> bool:
    try:
        logger.debug(f'moving "{source}" to "{target}"')
        os.replace(source, target)
        return True
    except Exception as e:
        message = f'failed to move file "{source}" to desired location "{target}": {e}'
        logger.warning(message)
        return False


@Plugins.register('cache', Plugins.kind.ACTOR_CONFIG)
class FileCacheConfig(FileDownloadConfig):
    pass


@Plugins.register('cache', Plugins.kind.ACTOR_ENTITY)
class FileCacheEntity(QueueActionEntity):
    url_fields: List[str] = ['attachments', 'thumbnail_url', 'avatar_url']
    """names of fields in the incoming record containing urls of files to be downloaded.
    Field must contain url, list of urls, or a nested record to look for fields names into"""
    replace_after: Optional[NonNegativeFloat] = None
    """how old existing file should be to get redownloaded, in hours"""
    consume_record: bool = False
    """whether record should be consumed or passed down the chain after processing"""
    import_path: Optional[str] = None
    """path to external location to look for a file before downloading. Supports templating with '{...}'"""
    import_filename: Optional[str] = None
    """name (without extension) of external file to use instead of downloading.
    If file exists, it is copied to the cache directory once for every processed url. Supports templating with '{...}'"""
    import_rename_suffix: str = ' [{i}]'
    """for fields containing list of urls, this suffix is added to import_filename template to look for additional files.
    Must contain {i} exactly once"""
    export_path: Optional[str] = None
    """path to external location to store a copy of cached file. Supports templating with '{...}'"""
    export_filename: Optional[str] = None
    """name used to store a copy of cached file externally. Supports templating with '{...}'"""
    export_rename_suffix: str = ' [{i}]'
    """if export_filename already exists, this suffix is used to generate a new, unique name.
    Must contain {i} exactly once"""

    @model_validator(mode='after')
    def check_paths(self):
        if (self.import_path and not self.import_filename) or (not self.import_path and self.import_filename):
            raise ValueError(f'both import_path and import_filename should be present if specified')
        if (self.export_path and not self.export_filename) or (not self.export_path and self.export_filename):
            raise ValueError(f'both export_path and export_filename should be present if specified')
        return self


@Plugins.register('cache', Plugins.kind.ACTOR)
class FileCacheAction(QueueAction):
    """
    Cache url locally

    For every incoming record, go through fields specified in "url_fields" setting
    and download files the urls are pointing to. Downloaded files are stored under
    the path, defined by the cache_directory parameter in application-wide settings,
    where they are used to present record in the web interface.

    It is possible to reuse files already stored by the "download" plugin with
    import_path/import_filename template options, though it might not work well
    with the "attachments" field. A copy of cached file can be
    stored to external location by providing export_path/export_filename templates.

    The way files are stored internally might change in the future, leaving already
    stored files inaccessible by the web interface. Use export_path/export_filename
    and import_path/import_filename options to define external persistent storage
    layout, that doesn't depend on the cache format.
    """

    def __init__(self, conf: FileCacheConfig, entities: Sequence[FileCacheEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.conf: FileCacheConfig
        self.entities: Mapping[str, FileCacheEntity]
        self.concurrency_limit = asyncio.BoundedSemaphore(value=conf.max_concurrent_downloads)
        settings: Optional[SettingsSection] = ctx.get_extra('settings')
        if settings is None:
            raise RuntimeError(f'runtime context is missing Settings instance. This is a bug, please report it')
        self.cache = FileCache(settings.cache_directory, self.conf.partial_file_suffix)

    async def handle_single_record(self, logger: logging.Logger, client: HttpClient,
                                   entity: FileCacheEntity, record: Record) -> None:
        for field in entity.url_fields:
            await self._handle_field(logger, client, entity, record, field)

    async def _handle_field(self, logger: logging.Logger, client: HttpClient,
                            entity: FileCacheEntity, record: Record, field_name: str):
        field = getattr(record, field_name, None)
        if field is None:
            logger.debug(f'no field "{field_name}" in record {record!r}, skipping')
        elif isinstance(field, str):
            await self._cache_urls(logger, client, entity, record, [field])
        elif isinstance(field, list):
            await self._cache_urls(logger, client, entity, record, field)
        elif isinstance(field, Record):
            await self.handle_single_record(logger, client, entity, field)
        else:
            msg = f'field "{field_name}" of record {record!r} does not seem to hold any links. Raw field value: {field}'
            logger.debug(msg)

    def _find_external_files(self, entity: FileCacheEntity, record: Record) -> List[Path]:
        if entity.import_path and entity.import_filename:
            external_path = Fmt.format_filename(entity.import_path, entity.import_filename, record)
            external_files = find_with_suffix(external_path, entity.import_rename_suffix)
            self.logger.debug(f'[{entity.name}] external files found: {len(external_files)}')
            return external_files
        return []

    def _export(self, entity: FileCacheEntity, record: Record, cache_path: Optional[Path]):
        if cache_path is None:
            return
        if not entity.export_path or not entity.export_filename:
            return
        export_path = Fmt.format_filename(entity.export_path, entity.export_filename, record)
        export_path = find_free_suffix(export_path, entity.export_rename_suffix)
        export_path = export_path.with_suffix(cache_path.suffix)
        try:
            self.logger.debug(f'[{entity.name}] creating a copy of "{cache_path}" at "{export_path}"')
            shutil.copy2(cache_path, export_path)
        except OSError as e:
            self.logger.warning(f'[{entity.name}] failed to copy "{cache_path}" to "{export_path}": {e}')

    async def _cache_urls(self, logger: logging.Logger, client: HttpClient,
                          entity: FileCacheEntity, record: Record, maybe_urls: List[str]) -> None:
        urls = []
        for url in maybe_urls:
            if is_url(url):
                urls.append(url)
            else:
                logger.debug(f'"{url}" does not seem to be a valid url, skipping. Record: {record!r}')

        external_files: Sequence[Optional[Path]] = self._find_external_files(entity, record)
        if not external_files or len(external_files) != len(urls):
            external_files = [None] * len(urls)

        for url, external_file in zip(urls, external_files):
            async with self.concurrency_limit:
                cache_path = await self.cache.store(logger, client, record, url, entity.replace_after, external_file)
            self._export(entity, record, cache_path)




import abc
import asyncio
import datetime
import json
import logging
import re
import time
import urllib.parse
from collections import defaultdict
from dataclasses import dataclass, field
from email.utils import parsedate_to_datetime
from http.cookies import SimpleCookie
from math import log2
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple, Union

import aiohttp
import multidict
from aiohttp.abc import AbstractCookieJar
from multidict import CIMultiDictProxy

from avtdl._version import __version__
from avtdl.core.utils import JSONType, convert_cookiejar, load_cookies, timeit, utcnow

HIGHEST_UPDATE_INTERVAL: float = 4000


@dataclass
class RetrySettings:
    retry_times: int = 0
    """transparent retrying: number of attempts"""
    retry_delay: float = 1
    """transparent retrying: delay before first retry attempt"""
    retry_multiplier: float = 1.2
    """transparent retrying: factor to increase retry delay compared to the previous attempt"""

    def __post_init__(self):
        if self.retry_times < 0:
            raise ValueError(f'retry_times must be positive, got "{self.retry_times}"')


@dataclass
class EndpointState:
    etag: Optional[str] = None
    last_modified: Optional[str] = None

    def update(self, headers: Union[CIMultiDictProxy, Dict[str, str]]):
        self.last_modified = headers.get('Last-Modified', None)
        self.etag = headers.get('Etag', None)


class StateStorage:
    """
    Store EndpointState objects grouped by (url, method, params) tuples
    """

    def __init__(self) -> None:
        self._storage: Dict[(Tuple[str, str, Optional[Any]]), EndpointState] = defaultdict(EndpointState)

    def get(self, url: str, method: str, params: Optional[Any]) -> EndpointState:
        encoded_params = urllib.parse.urlencode(sorted(params.items())) if params is not None else None
        return self._storage[(url, method, encoded_params)]


class Delay:
    """Provide method to calculate next delay for exponential backoff based on S-shaped curve"""

    A: float = 4600  # upper asymptote
    k: float = 0.88  # curve growth rate
    x0: float = 8  # x value corresponding to midpoint of the curve

    @classmethod
    def _sigmoid(cls, x: float) -> float:
        y = cls.A / (1 + 2 ** (-cls.k * (x - cls.x0)))
        return y

    @classmethod
    def _inv_sigmoid(cls, y: float) -> float:
        if y <= 0:
            return 0
        # raises ValueError if y >= cls.A
        x = cls.x0 - log2((cls.A - y) / y) / cls.k
        return x

    @classmethod
    def get_next(cls, current: float) -> float:
        """
        Find current value on S-shaped curve and return a next one

        Returns 0 if input is not higher than 0,
        returns cls.A if input higher than cls.A.
        """
        try:
            current_step = cls._inv_sigmoid(current)
        except ValueError:
            current_step = current
        next_step = current_step + 1
        next_delay = cls._sigmoid(next_step)
        return next_delay


def get_cache_ttl(headers: Union[multidict.CIMultiDictProxy, Dict[str, str]]) -> Optional[int]:
    """
    check for Expires and Cache-Control headers, return integer representing
    how many seconds is left until resource is outdated, if they are present
    and was parsed successfully
    """

    def get_expires_from_cache_control(headers: Union[multidict.CIMultiDictProxy, Dict[str, str]]) -> Optional[datetime.datetime]:
        cache_control = headers.get('Cache-Control', '')
        if 'must-revalidate' in cache_control.lower():
            return None
        try:
            [max_age] = re.findall('max-age=(\d+)', cache_control, re.IGNORECASE)
            max_age_value = datetime.timedelta(seconds=int(max_age))
        except (IndexError, TypeError, ValueError):
            return None

        try:
            date_value = parsedate_to_datetime(headers.get('Date', ''))
        except (TypeError, ValueError):
            date_value = None
        try:
            last_modified_value = parsedate_to_datetime(headers.get('Last-Modified', ''))
        except (TypeError, ValueError):
            last_modified_value = None

        calculated_update_date = date_value or last_modified_value or utcnow()
        expires = calculated_update_date + max_age_value
        return expires

    def get_expires_from_expires_header(headers: Union[multidict.CIMultiDictProxy, Dict[str, str]]) -> Optional[datetime.datetime]:
        try:
            expires_header = headers.get('Expires')
            if expires_header is None or expires_header == '0':
                return None
            expires_value = parsedate_to_datetime(expires_header)
            return expires_value
        except (TypeError, ValueError):
            return None

    expires = get_expires_from_cache_control(headers) or get_expires_from_expires_header(headers)
    if expires is None:
        return None

    delta = (expires - utcnow()).total_seconds()
    if delta < 0:
        return None

    return int(delta)


def get_retry_after(headers: Union[Dict[str, str], multidict.CIMultiDictProxy[str]]) -> Optional[int]:
    """return parsed value of Retry-After header, if present"""
    retry_after = headers.get('Retry-After')
    if retry_after is None:
        return None
    try:
        return int(retry_after)
    except ValueError:
        pass
    try:
        retry_at = parsedate_to_datetime(retry_after)
        delay = int((retry_at - utcnow()).total_seconds())
        if delay > 0:
            return delay
    except (TypeError, ValueError):
        pass
    return None


def insert_useragent(headers: Optional[Dict[str, str]]) -> Optional[Dict[str, str]]:
    if headers is None:
        return None
    if 'User-Agent' in headers:
        return headers
    headers = headers.copy()
    headers['User-Agent'] = f'avtdl {__version__}'
    return headers


@dataclass
class NoResponse:
    logger: logging.Logger
    e: Exception
    url: str
    text: None = None
    completed: bool = False
    ok: bool = False
    has_content: bool = False
    status: int = 0
    headers: None = None

    def has_json(self) -> bool:
        return False

    def json(self) -> JSONType:
        return None

    def next_update_interval(self, base: float, current: float, adjust_update_interval: bool = True) -> float:
        return decide_on_update_interval(self.logger, self.url, None, None, current, base, adjust_update_interval)


@dataclass
class HttpResponse:
    logger: logging.Logger
    text: str
    """response body in plaintext"""
    url: str
    """request url"""
    ok: bool
    """status is not 4xx or 5xx"""
    has_content: bool
    """status is 2xx"""
    status: int
    reason: str
    headers: Union[CIMultiDictProxy[str], Dict[str, str]]
    """response headers"""
    request_headers: Union[CIMultiDictProxy[str], Dict[str, str]]
    cookies: SimpleCookie
    """response cookies"""
    endpoint_state: EndpointState
    content_encoding: str
    """detected response content encoding"""
    completed: bool = True
    _json: Optional[JSONType] = None

    @classmethod
    def from_response(cls, response: aiohttp.ClientResponse, text: str, state: EndpointState, logger: logging.Logger):
        has_content = 200 <= response.status < 300
        factory: type[HttpResponse]
        if has_content:
            factory = DataResponse
        elif response.ok:
            factory = GoodResponse
        elif not response.ok:
            factory = BadResponse
        else:
            factory = cls
        response = factory(
            logger,
            text,
            str(response.url),
            response.ok,
            has_content,
            response.status,
            response.reason or 'No reason',
            response.headers,
            response.request_info.headers,
            response.cookies,
            state,
            response.get_encoding()
        )
        return response

    def has_json(self) -> bool:
        try:
            _ = self.json()
            return True
        except json.JSONDecodeError:
            return False

    def json(self) -> JSONType:
        if self._json is None:
            self._json = json.loads(self.text)
        return self._json

    def next_update_interval(self, base: float, current: float, adjust_update_interval: bool = True) -> float:
        return decide_on_update_interval(self.logger, self.url, self.status, self.headers, current, base,
                                         adjust_update_interval)


MaybeHttpResponse = Union[NoResponse, HttpResponse]


class BadResponse(HttpResponse):
    """HttpResponse that completed with error (status >= 400)"""
    ok: Literal[False] = False


class GoodResponse(HttpResponse):
    """HttpResponse that completed successfully (status < 400)"""
    ok: Literal[True] = True


class DataResponse(GoodResponse):
    """HttpResponse that completed successfully"""
    has_content: Literal[True] = True


class RateLimit:
    """
    Encapsulate state of rate limits for endpoint

    >>> async def endpoint_request(url: str, client: HttpClient, rate_limit: RateLimit):
    >>>     async with rate_limit:
    >>>         response = await client.request(url)
    >>>         rate_limit.submit_response(response)
    >>>

    Base rate limit, taking into account response status and RetryAfter header
    """

    DEFAULT_DELAY = 10

    def __init__(self, name: str, logger: Optional[logging.Logger] = None, base_delay: int = DEFAULT_DELAY) -> None:
        self.ready_at: datetime.datetime = utcnow()
        self.base_delay = base_delay
        self.current_delay = base_delay

        self.name = name
        self.logger = logger or logging.getLogger().getChild('rate_limit')
        self.lock = asyncio.Lock()
        self.perf_lock_acquired_at: float = 0

    async def __aenter__(self) -> 'RateLimit':
        with timeit() as t:
            await self.lock.acquire()
        self.perf_lock_acquired_at = t.end
        if self.delay > 0:
            self.logger.debug(
                f'[{self.name}] lock acquired in {t.timedelta}, {self.delay} seconds until rate limit reset')
            await asyncio.sleep(self.delay)
        else:
            self.logger.debug(f'[{self.name}] lock acquired in {t.timedelta}, there is no active rate limit')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()
        duration = datetime.timedelta(seconds=(time.perf_counter() - self.perf_lock_acquired_at))
        self.logger.debug(f'[{self.name}] lock released after {duration}')
        return False

    @property
    def delay(self) -> int:
        """seconds left until next request can be made"""
        delay = int((self.ready_at - utcnow()).total_seconds()) + 1
        reset_after = max(0, delay)
        return reset_after

    def submit_response(self, response: MaybeHttpResponse, logger: Optional[logging.Logger] = None):
        """
        Update limits values and reset time using data from response.

        Client code should call it after completing request. Client code might
        provide a custom Logger instance to help with identifying debug output
        from specific request.
        """
        logger = logger or self.logger
        self.current_delay = self._submit_response(response, logger)
        self.ready_at = (utcnow() + datetime.timedelta(seconds=self.current_delay))

    @abc.abstractmethod
    def _submit_response(self, response: MaybeHttpResponse, logger: logging.Logger) -> int:
        """parse response and return minimum delay until the next request, in seconds"""


class HttpRateLimit(RateLimit):

    def _submit_response(self, response: MaybeHttpResponse, logger: logging.Logger) -> int:
        return int(response.next_update_interval(self.base_delay, self.current_delay))


class BucketRateLimit(RateLimit):
    """Rate limit for token bucket type of endpoint"""

    def __init__(self, name: str, logger: Optional[logging.Logger] = None) -> None:
        super().__init__(name, logger)
        self.limit_total: int = 50
        self.limit_remaining: int = 10
        self.reset_at: int = int(utcnow().timestamp())

    def _fallback_delay(self, response: MaybeHttpResponse, logger: logging.Logger) -> int:
        return int(response.next_update_interval(self.base_delay, self.current_delay, True))

    def _submit_response(self, response: MaybeHttpResponse, logger: logging.Logger) -> int:
        logger = logger or self.logger

        if isinstance(response, NoResponse):
            return self._fallback_delay(response, logger)
        parsed_successfully = self._submit_headers(response, logger)
        if not parsed_successfully:
            return self._fallback_delay(response, logger)
        if self.limit_remaining >= 1:
            if response.ok:
                return 0
            return self._fallback_delay(response, logger)
        reset_after = max(0, self.reset_at - int(utcnow().timestamp()))
        return reset_after

    @abc.abstractmethod
    def _submit_headers(self, response: HttpResponse, logger: logging.Logger) -> bool:
        """
        Parse response headers and update self.limit_total, self.limit_remaining and self.reset_at,
        returns True if headers are present and successfully parsed
        """


class NoRateLimit(RateLimit):

    def __init__(self, name: str, logger: Optional[logging.Logger] = None) -> None:
        logger = logging.getLogger('noop')
        logger.setLevel(logging.CRITICAL)
        super().__init__(name, logger)

    def _submit_response(self, response: MaybeHttpResponse, logger: logging.Logger) -> float:
        return 0


@dataclass
class RequestDetails:
    """Represents parameters of a request to be made to specific endpoint"""
    url: str
    method: str = 'GET'
    params: Optional[Dict[str, Any]] = None
    data: Optional[Any] = None
    data_json: Optional[JSONType] = None
    headers: Optional[Dict[str, Any]] = None
    rate_limit: RateLimit = field(default_factory=lambda: NoRateLimit('default'))
    endpoint_state: EndpointState = field(default_factory=EndpointState)
    retry_settings: RetrySettings = field(default_factory=RetrySettings)


class Endpoint(abc.ABC):
    """
    Superclass providing utility methods for concrete Endpoints

    Concrete Endpoints must implement prepare() method taking arbitrary
    arguments, that returns RequestDetails instance.
    """

    @abc.abstractmethod
    def prepare(self, *args, **kwargs) -> RequestDetails:
        """Prepare a RequestDetails object based on passed arguments"""


def decide_on_update_interval(logger: logging.Logger, url: str, status: Optional[int],
                              headers: Union[CIMultiDictProxy[str], Dict[str, str], None],
                              current_update_interval: float, base_update_interval: float,
                              adjust_update_interval: bool = True) -> float:
    update_interval: float

    if status is None or headers is None:  # response hasn't completed due to network error
        update_interval = max(Delay.get_next(current_update_interval), current_update_interval)
        logger.warning(f'update interval set to {update_interval} seconds for {url}')
        return update_interval

    retry_after = get_retry_after(headers)
    if retry_after is not None:
        raw_header = headers.get("Retry-After")
        logger.debug(f'got Retry-After header with value {raw_header}')
        update_interval = min(float(retry_after), HIGHEST_UPDATE_INTERVAL)
        if update_interval < base_update_interval:
            update_interval = base_update_interval
        logger.warning(f'update interval set to {update_interval} seconds for {url} as requested by Retry-After')
    elif status >= 400:
        update_interval = max(Delay.get_next(current_update_interval), current_update_interval)
        logger.warning(f'update interval set to {update_interval} seconds for {url}')
    else:
        if adjust_update_interval:
            new_update_interval = get_cache_ttl(headers) or base_update_interval
            new_update_interval = min(new_update_interval, 10 * base_update_interval)  # in case ttl is overly long
            new_update_interval = max(new_update_interval, base_update_interval)
            if new_update_interval != current_update_interval:
                logger.info(f'next update in {new_update_interval}')
            update_interval = new_update_interval
        else:
            if current_update_interval != base_update_interval:
                logger.info(f'restoring update interval {base_update_interval} seconds for {url}')
            update_interval = base_update_interval

    return update_interval


class HttpClient:

    def __init__(self, logger: logging.Logger, session: aiohttp.ClientSession):
        self.logger = logger
        self.session = session

    @property
    def cookie_jar(self) -> AbstractCookieJar:
        return self.session.cookie_jar

    async def request_once(self, url: str,
                           params: Optional[Dict[str, str]] = None,
                           data: Optional[Any] = None,
                           data_json: Optional[Any] = None,
                           headers: Optional[Dict[str, Any]] = None,
                           method: str = 'GET',
                           state: EndpointState = EndpointState(),
                           ) -> 'MaybeHttpResponse':
        logger = self.logger

        request_headers: Dict[str, Any] = headers or {}
        if self.session.headers is not None:
            request_headers.update(self.session.headers)
        if state.last_modified is not None and method in ['GET', 'HEAD']:
            request_headers['If-Modified-Since'] = state.last_modified
        if state.etag is not None:
            request_headers['If-None-Match'] = state.etag
        request_headers = insert_useragent(request_headers)
        server_hostname = self.session.headers.get('Host') or request_headers.get('Host') or None
        try:
            async with self.session.request(method, url, headers=request_headers, params=params, data=data,
                                            json=data_json, server_hostname=server_hostname) as client_response:
                # fully read http response to get it cached inside ClientResponse object
                # client code can then use it by awaiting .text() again without causing
                # network activity and potentially triggering associated errors
                text = await client_response.text()
        except Exception as e:
            logger.warning(f'error while fetching {url}: {e.__class__.__name__} {e}')
            return NoResponse(logger, e, url)

        if not client_response.ok:
            logger.warning(
                f'got code {client_response.status} ({client_response.reason or "No reason"}) while fetching {url}')
            if text:
                logger.debug(f'response body: "{text}"')
        elif client_response.status != 304:
            # some servers do not have cache headers in 304 response, so only updating on 200
            state.update(client_response.headers)

            cache_control = client_response.headers.get('Cache-control')
            logger.debug(
                f'Last-Modified={state.last_modified or "absent"}, ETAG={state.etag or "absent"}, Cache-control="{cache_control or "absent"}" for {client_response.real_url}')

        response = HttpResponse.from_response(client_response, text, state, logger)
        return response

    async def request(self, url: str,
                      params: Optional[Dict[str, str]] = None,
                      data: Optional[Any] = None,
                      data_json: Optional[Any] = None,
                      headers: Optional[Dict[str, Any]] = None,
                      method: str = 'GET',
                      state: EndpointState = EndpointState(),
                      settings: RetrySettings = RetrySettings()) -> 'MaybeHttpResponse':
        response: MaybeHttpResponse = NoResponse(self.logger, Exception('request_once was never called'), url)
        next_try_delay = settings.retry_delay
        for attempt in range(settings.retry_times + 1):
            response = await self.request_once(url, params, data, data_json, headers, method, state)
            if response is not None and response.ok:
                break
            next_try_delay *= settings.retry_multiplier
        return response

    async def request_text(self, url: str,
                           params: Optional[Dict[str, str]] = None,
                           data: Optional[Any] = None,
                           data_json: Optional[Any] = None,
                           headers: Optional[Dict[str, Any]] = None,
                           method: str = 'GET',
                           state: EndpointState = EndpointState(),
                           settings: RetrySettings = RetrySettings()) -> Optional[str]:
        response = await self.request(url, params, data, data_json, headers, method, state, settings)
        if isinstance(response, DataResponse):
            return response.text
        return None

    async def request_json(self, url: str,
                           params: Optional[Dict[str, str]] = None,
                           data: Optional[Any] = None,
                           data_json: Optional[Any] = None,
                           headers: Optional[Dict[str, Any]] = None,
                           method: str = 'GET',
                           state: EndpointState = EndpointState(),
                           settings: RetrySettings = RetrySettings()) -> Optional[JSONType]:
        response = await self.request(url, params, data, data_json, headers, method, state, settings)
        if response.has_json():
            return response.json()
        return None

    async def request_endpoint(self, logger: logging.Logger, details: RequestDetails) -> MaybeHttpResponse:
        async with details.rate_limit:
            response = await self.request(url=details.url,
                                          params=details.params,
                                          data=details.data,
                                          data_json=details.data_json,
                                          headers=details.headers,
                                          method=details.method,
                                          state=details.endpoint_state,
                                          settings=details.retry_settings)
            details.rate_limit.submit_response(response, logger)
        return response


class SessionStorage:
    """
    Provide way to initialize, store and reuse ClientSession objects

    The "name" parameter can be used to get distinct sessions
    with the same cookies and headers.

    To ensure sessions shared between multiple tasks are safely closed,
    user must call SessionStorage.run() once from a running event loop
    and should not use "async with" on any of the session objects, as it
    might cause the session to be closed prematurely, making other tasks
    using it fail.
    """

    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.sessions: Dict[str, aiohttp.ClientSession] = {}
        self.task: Optional[asyncio.Task] = None
        self.logger = (logger or logging.getLogger()).getChild('sessions')

    @staticmethod
    def get_session_id(cookies_file: Optional[Path], headers: Optional[Dict[str, Any]], name: str = '') -> str:
        return name + str((cookies_file, headers))

    def get_session_by_id(self, session_id: str) -> Optional[aiohttp.ClientSession]:
        return self.sessions.get(session_id)

    def session_exists(self, cookies_file: Optional[Path], headers: Optional[Dict[str, Any]], name: str = '') -> bool:
        session_id = self.get_session_id(cookies_file, headers, name)
        session = self.get_session_by_id(session_id)
        return session is not None

    def get_session(self, cookies_file: Optional[Path] = None, headers: Optional[Dict[str, Any]] = None,
                    name: str = '') -> aiohttp.ClientSession:
        session_id = self.get_session_id(cookies_file, headers, name)
        session = self.get_session_by_id(session_id)
        if session is None:
            netscape_cookies = load_cookies(cookies_file)
            cookies = convert_cookiejar(netscape_cookies) if netscape_cookies else None
            session = aiohttp.ClientSession(cookie_jar=cookies, headers=headers)
            self.sessions[session_id] = session
        return session

    async def ensure_closed(self) -> None:
        try:
            await asyncio.Future()
        except (asyncio.CancelledError, KeyboardInterrupt):
            await self.close()

    async def close(self) -> None:
        self.logger.debug('closing http sessions...')
        for session_id, session in self.sessions.items():
            if not session.closed:
                self.logger.debug(f'closing session "{session_id}"')
                await session.close()
        self.logger.debug('done')



import logging
import pickle
from pathlib import Path
from typing import Optional, Protocol, Type, TypeVar

from pydantic import BaseModel, ValidationError

from avtdl.core.utils import check_dir, format_validation_error


class StateSerializable(Protocol):
    """
    Protocol to allow storing and restoring object state on disk

    The directory argument provides path to the directory used to
    store data, leaving the filename to be decided by the implementations.

    Implementations must not raise errors related to file storing,
    reading and parsing, but may warn user about it.
    """

    def dump_state(self, directory: Path):
        """Serialize the current state and write it to directory"""

    def apply_state(self, directory: Path):
        """Load serialized state from directory and apply it to the object"""


DataType = TypeVar('DataType', bound=BaseModel)


class StateSerializer:
    """Utility class for handling storing and loading state"""
    logger = logging.getLogger().getChild('state_storage')

    @staticmethod
    def serialize(obj: object) -> bytes:
        return pickle.dumps(obj)

    @classmethod
    def dump(cls, obj: DataType, path: Path) -> bool:
        """Store state to file"""
        try:
            cls.logger.debug(f'storing runtime state to {path}')
            data = cls.serialize(obj)
        except Exception as e:
            cls.logger.warning(f'failed to serialize state to "{path}": {e}')
            cls.logger.debug(f'object that failed to serialize: {obj}', exc_info=True)
            return False
        try:
            check_dir(path.parent)
            path.write_bytes(data)
            return True
        except OSError as e:
            cls.logger.warning(f'failed to store state to "{path}": {e}')
            return False

    @staticmethod
    def deserialize(data: bytes):
        return pickle.loads(data)

    @classmethod
    def restore(cls, Model: Type[DataType], path: Path) -> Optional[DataType]:
        """Load state from file and apply to the object"""
        if not path.exists():
            cls.logger.debug(f'skipping restore: "{path}" does not exist')
            return None
        cls.logger.info(f'restoring runtime state from {path}')
        try:
            raw_data = path.read_bytes()
        except (OSError, UnicodeDecodeError) as e:
            cls.logger.warning(f'failed to read state from "{path}": {e}')
            return None
        try:
            obj = cls.deserialize(raw_data)
            return Model.model_validate(obj)
        except pickle.UnpicklingError as e:
            cls.logger.warning(f'failed to load state from "{path}": {e}')
            return None
        except ValidationError as e:
            msg = f'failed to parse state from "{path}": '
            cls.logger.warning(format_validation_error(e, msg))
            return None
        except Exception as e:
            cls.logger.warning(f'error restoring state from "{path}": {e}')
            return None



import datetime
import logging
from http.cookies import SimpleCookie
from typing import Callable, Dict, List

import pytest

import avtdl
from avtdl.core.request import Delay, EndpointState, HttpRateLimit, HttpResponse, MaybeHttpResponse, NoResponse, \
    RateLimit
from avtdl.plugins.discord.webhook import DiscordRateLimit
from avtdl.plugins.twitter.endpoints import TwitterRateLimit

BASE_DELAY = 600
MAX_AGE_DELAY = 900
RETRY_AFTER_DELAY = 1200
SHORT_RETRY_AFTER_DELAY = 12
BASE_BACKOFF_DELAY = int(Delay.get_next(BASE_DELAY))
NEXT_BASE_BACKOFF_DELAY = int(Delay.get_next(BASE_BACKOFF_DELAY))

TIMESTAMP_NOW = 1750000000
RATE_LIMIT_DELAY = 1000
RATE_LIMIT_RESET = TIMESTAMP_NOW + RATE_LIMIT_DELAY
DEFAULT_BACKOFF_DELAY = int(Delay.get_next(RateLimit.DEFAULT_DELAY))
NEXT_DEFAULT_BACKOFF_DELAY = int(Delay.get_next(DEFAULT_BACKOFF_DELAY))


def mocked_utcnow():
    return datetime.datetime.fromtimestamp(TIMESTAMP_NOW, tz=datetime.timezone.utc)


@pytest.fixture
def mock_utcnow(monkeypatch):
    monkeypatch.setattr(avtdl.core.request, 'utcnow', mocked_utcnow)


def prepare_response(status: int, headers: Dict[str, str]) -> HttpResponse:
    endpoint_state = EndpointState()
    endpoint_state.update(headers)
    return HttpResponse(
        logging.getLogger('test'),
        text='test body',
        url='https://example.com',
        ok=status < 400,
        has_content=status < 300,
        status=status,
        reason='No reason provided',
        headers=headers,
        request_headers={},
        cookies=SimpleCookie(),
        endpoint_state=endpoint_state,
        content_encoding='utf8'
    )


def prepare_no_response() -> MaybeHttpResponse:
    return NoResponse(
        logging.getLogger('test'),
        e=Exception('network error'),
        url='https://example.com'
    )


def check_rate_limit_delay(rate_limit: RateLimit, responses: List[HttpResponse]) -> int:
    """Apply responses to rate_limit and return resulting current_delay"""
    for response in responses:
        rate_limit.submit_response(response, response.logger)
    return rate_limit.delay - 1


def check_for_responses(rate_limit: RateLimit, responses: List[Callable[[], HttpResponse]], expected_delay: int):
    response_instances = [response_factory() for response_factory in responses]
    actual_delay = check_rate_limit_delay(rate_limit, response_instances)
    if actual_delay != expected_delay:
        actual_delay = check_rate_limit_delay(rate_limit, response_instances)
    assert actual_delay == expected_delay


def prepare_200() -> HttpResponse:
    return prepare_response(200, {})


def prepare_200_with_max_age() -> HttpResponse:
    return prepare_response(200, {'Cache-Control': f'max-age={MAX_AGE_DELAY}'})


def prepare_304() -> HttpResponse:
    return prepare_response(304, {})


def prepare_304_with_max_age() -> HttpResponse:
    return prepare_response(304, {'Cache-Control': f'max-age={MAX_AGE_DELAY}'})


def prepare_403() -> HttpResponse:
    return prepare_response(403, {})


def prepare_429() -> HttpResponse:
    return prepare_response(429, {'Retry-After': str(RETRY_AFTER_DELAY)})

def prepare_short_429() -> HttpResponse:
    return prepare_response(429, {'Retry-After': str(SHORT_RETRY_AFTER_DELAY)})


def prepare_500() -> HttpResponse:
    return prepare_response(500, {})


class TestHttpRateLimit:
    testcases = {
        '200': ([prepare_200], BASE_DELAY),
        '200 with max age': ([prepare_200_with_max_age], MAX_AGE_DELAY),
        'consecutive 200': ([prepare_200, prepare_200], BASE_DELAY),
        '304': ([prepare_304], BASE_DELAY),
        '304 with max age': ([prepare_304_with_max_age], MAX_AGE_DELAY),
        '304 does not change delay': ([prepare_200, prepare_304], BASE_DELAY),
        'consecutive 304': ([prepare_304, prepare_304], BASE_DELAY),
        '429': ([prepare_429], RETRY_AFTER_DELAY),
        '429 with short Retry-After': ([prepare_short_429], BASE_DELAY),
        '500': ([prepare_500], BASE_BACKOFF_DELAY),
        '200 after 500': ([prepare_500, prepare_200], BASE_DELAY),
        '304 after 500': ([prepare_500, prepare_304], BASE_DELAY),
        'consecutive error responses': ([prepare_500, prepare_403], NEXT_BASE_BACKOFF_DELAY),
        'network error': ([prepare_no_response], BASE_BACKOFF_DELAY),
        '200 after network error': ([prepare_no_response, prepare_200], BASE_DELAY),
        '304 after network error': ([prepare_no_response, prepare_304], BASE_DELAY),
        '500 after network error': ([prepare_no_response, prepare_500], NEXT_BASE_BACKOFF_DELAY),
        'consecutive network errors': ([prepare_no_response, prepare_no_response], NEXT_BASE_BACKOFF_DELAY),
    }

    @staticmethod
    def rate_limit() -> RateLimit:
        return HttpRateLimit(name='test', base_delay=BASE_DELAY)

    @pytest.mark.parametrize('responses, expected_delay', testcases.values(), ids=testcases.keys())
    def test(self, mock_utcnow, responses: List[Callable[[], HttpResponse]], expected_delay: int):
        check_for_responses(self.rate_limit(), responses, expected_delay)


class TestLongHttpRateLimit:
    testcases = {
        'limit longer than max-age is not adjusted': ([prepare_200_with_max_age], 14400),
        'long limit not affected by retry-after': ([prepare_429], 14400),
        'limit longer than HIGHEST_UPDATE_INTERVAL is not adjusted on bad response': ([prepare_403], 14400),
        'limit longer than HIGHEST_UPDATE_INTERVAL is not adjusted on network error': ([prepare_no_response], 14400),
    }

    @staticmethod
    def rate_limit() -> RateLimit:
        return HttpRateLimit(name='test', base_delay=14400)

    @pytest.mark.parametrize('responses, expected_delay', testcases.values(), ids=testcases.keys())
    def test(self, mock_utcnow, responses: List[Callable[[], HttpResponse]], expected_delay: int):
        check_for_responses(self.rate_limit(), responses, expected_delay)


def prepare_twitter_bucket(status: int, full: bool, reset_at: int) -> HttpResponse:
    return prepare_response(status, {
        'x-rate-limit-limit': '50',
        'x-rate-limit-remaining': '48' if full else '0',
        'x-rate-limit-reset': str(reset_at)
    })


def twitter_200_bucket_full() -> HttpResponse:
    return prepare_twitter_bucket(200, True, RATE_LIMIT_RESET)


def twitter_200_bucket_empty() -> HttpResponse:
    return prepare_twitter_bucket(200, False, RATE_LIMIT_RESET)


def twitter_404_bucket_full() -> HttpResponse:
    return prepare_twitter_bucket(404, True, RATE_LIMIT_RESET)


def twitter_404_bucket_empty() -> HttpResponse:
    return prepare_twitter_bucket(404, False, RATE_LIMIT_RESET)


class TestTwitterRateLimit:
    testcases = {
        'success response with limit available': ([twitter_200_bucket_full], 0),
        'success response with limit reached': ([twitter_200_bucket_empty], RATE_LIMIT_DELAY),
        'error response with limit available': ([twitter_404_bucket_full], DEFAULT_BACKOFF_DELAY),
        'error response with limit reached': ([twitter_404_bucket_empty], RATE_LIMIT_DELAY),
        '429 with Retry-After and no limit headers': ([prepare_429], RETRY_AFTER_DELAY),
        '500 without limit headers': ([prepare_500], DEFAULT_BACKOFF_DELAY),
        'network error': ([prepare_no_response], DEFAULT_BACKOFF_DELAY),
        '200 after 500': ([prepare_500, twitter_200_bucket_full], 0),
        'consecutive 500': ([prepare_500, prepare_500], NEXT_DEFAULT_BACKOFF_DELAY),
        '200 after network error': ([prepare_no_response, twitter_200_bucket_full], 0),
        '500 after network error': ([prepare_no_response, prepare_500], NEXT_DEFAULT_BACKOFF_DELAY),
        'consecutive network errors': ([prepare_no_response, prepare_no_response], NEXT_DEFAULT_BACKOFF_DELAY),
    }

    @staticmethod
    def rate_limit() -> RateLimit:
        return TwitterRateLimit(name='test')

    @pytest.mark.parametrize('responses, expected_delay', testcases.values(), ids=testcases.keys())
    def test(self, mock_utcnow, responses: List[Callable[[], HttpResponse]], expected_delay: int):
        check_for_responses(self.rate_limit(), responses, expected_delay)


def prepare_discord_bucket(status: int, full: bool, reset_at: int, bucket: str = 'test') -> HttpResponse:
    return prepare_response(status, {
        'X-RateLimit-Limit': '5',
        'X-RateLimit-Remaining': '4' if full else '0',
        'X-RateLimit-Reset': str(reset_at),
        'X-RateLimit-Bucket': bucket
    })
def discord_200_bucket_full() -> HttpResponse:
    return prepare_discord_bucket(200, True, RATE_LIMIT_RESET)


def discord_200_bucket_empty() -> HttpResponse:
    return prepare_discord_bucket(200, False, RATE_LIMIT_RESET)



class TestDiscordRateLimit:
    testcases = {
        'success response with limit available': ([discord_200_bucket_full], 0),
        'success response with limit reached': ([discord_200_bucket_empty], RATE_LIMIT_DELAY),
        '429 with Retry-After and no limit headers': ([prepare_429], RETRY_AFTER_DELAY),
        '500 without limit headers': ([prepare_500], DEFAULT_BACKOFF_DELAY),
        'network error': ([prepare_no_response], DEFAULT_BACKOFF_DELAY),
        '200 after 500': ([prepare_500, discord_200_bucket_full], 0),
        'consecutive 500': ([prepare_500, prepare_500], NEXT_DEFAULT_BACKOFF_DELAY),
        '200 after network error': ([prepare_no_response, discord_200_bucket_full], 0),
        '500 after network error': ([prepare_no_response, prepare_500], NEXT_DEFAULT_BACKOFF_DELAY),
        'consecutive network errors': ([prepare_no_response, prepare_no_response], NEXT_DEFAULT_BACKOFF_DELAY),
    }

    @staticmethod
    def rate_limit() -> RateLimit:
        return DiscordRateLimit(name='test')

    @pytest.mark.parametrize('responses, expected_delay', testcases.values(), ids=testcases.keys())
    def test(self, mock_utcnow, responses: List[Callable[[], HttpResponse]], expected_delay: int):
        check_for_responses(self.rate_limit(), responses, expected_delay)



import asyncio
import datetime
import json
import logging
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import multidict

from avtdl.core.actors import Action, ActionEntity, ActorConfig
from avtdl.core.formatters import DiscordEmbedLimits, MessageFormatter
from avtdl.core.interfaces import Record
from avtdl.core.plugins import Plugins
from avtdl.core.request import BucketRateLimit, Endpoint, HttpClient, HttpResponse, NoResponse, RequestDetails, \
    SessionStorage
from avtdl.core.runtime import RuntimeContext, TaskStatus


class DiscordRateLimit(BucketRateLimit):

    def __init__(self, name: str, on_submit: Callable[[HttpResponse], None] = lambda _: None):
        self.on_submit = on_submit
        self.bucket: Optional[str] = None
        super().__init__(name)

    def _submit_headers(self, response: HttpResponse, logger: logging.Logger) -> bool:
        headers = response.headers
        try:
            # "default" argument is set to 'NaN' to trigger ValueError, since None is disliked by typechecker
            self.limit_total = int(headers.get('X-RateLimit-Limit', 'NaN'))
            self.limit_remaining = int(headers.get('X-RateLimit-Remaining', 'NaN'))
            self.reset_at = int(headers.get('X-RateLimit-Reset', 'NaN'))
            self.bucket = headers.get('X-RateLimit-Bucket', None)
            ok = True
        except ValueError:
            logger.warning(f'[{self.name}] error parsing rate limit headers: "{headers}"')
            ok = False
        logger.debug(f'[{self.name}] rate limit {self.limit_remaining}/{self.limit_total}, resets after {datetime.timedelta(seconds=self.delay)}')
        self.on_submit(response)
        return ok

    @staticmethod
    def get_bucket(headers: Union[multidict.CIMultiDictProxy[str], Dict[str, str]]) -> Optional[str]:
        return headers.get('X-RateLimit-Bucket')


class WebhookEndpoint(Endpoint):

    def __init__(self) -> None:
        self.buckets: Dict[Optional[str], DiscordRateLimit] = {}
        self.buckets[None] = DiscordRateLimit('fallback bucket', on_submit=self.on_submit)
        self.urls_buckets: Dict[str, Optional[str]] = {}

    def on_submit(self, response: HttpResponse):
        bucket = DiscordRateLimit.get_bucket(response.headers)
        self.urls_buckets[response.url] = bucket
        if bucket not in self.buckets:
            name = bucket or 'fallback bucket'
            self.buckets[bucket] = DiscordRateLimit(name, on_submit=self.on_submit)

    def prepare(self, url, message: dict) -> RequestDetails:
        bucket = self.urls_buckets.get(url, None)
        rate_limit = self.buckets.get(bucket)
        if rate_limit is None:
            rate_limit = self.buckets[None]
            rate_limit.logger.warning(f'bucket "{bucket}" is associated with url "{url}" but does not have RateLimit')
        return RequestDetails(url, 'POST', data_json=message, rate_limit=rate_limit)


@Plugins.register('discord.hook', Plugins.kind.ACTOR_CONFIG)
class DiscordHookConfig(ActorConfig):
    pass


@Plugins.register('discord.hook', Plugins.kind.ACTOR_ENTITY)
class DiscordHookEntity(ActionEntity):
    url: str
    """webhook url"""


@Plugins.register('discord.hook', Plugins.kind.ACTOR)
class DiscordHook(Action):
    """
    Send record to Discord using webhook

    To generate webhook url follow instructions in "Making a Webhook" section of
    <https://support.discord.com/hc/en-us/articles/228383668-Intro-to-Webhooks>

    Some record types support rich formatting when sent to Discord, such as
    showing the author's avatar and links to attached images. Youtube videos will
    show thumbnail, however embedding video itself is not supported.

    Records coming within six seconds one after another will be batched together into a single message.
    When too many records are received at once, they will be sent with delays to conform to Discord
    rate limits. Records deemed to be too long to fit in a Discord message
    [length limits](https://discord.com/developers/docs/resources/channel#create-message-jsonform-params)
    will be dropped with a warning.
    """

    def __init__(self, conf: DiscordHookConfig, entities: Sequence[DiscordHookEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.sessions: SessionStorage = SessionStorage(self.logger)
        self.queues: Dict[str, asyncio.Queue] = {entity.name: asyncio.Queue() for entity in entities}
        self.endpoint = WebhookEndpoint()

    def handle(self, entity: DiscordHookEntity, record: Record):
        if not entity.name in self.queues:
            return
        self.logger.debug(f'[{entity.name}] adding record to send query: {record!r}')
        self.queues[entity.name].put_nowait(record)

    async def run(self):
        session = self.sessions.get_session(name=self.conf.name)
        client = HttpClient(self.logger, session)
        for entity in self.entities.values():
            name = f'{self.conf.name}:{entity.name}'
            info = TaskStatus(self.conf.name, entity.name)
            _ = self.controller.create_task(self.run_for(entity, client, info), name=name, _info=info)
        await self.sessions.ensure_closed()

    def update_status(self, info: TaskStatus, queue: asyncio.Queue, pending: List[Record]):
        msg = ''
        record = None
        if queue.qsize():
            msg += f'{queue.qsize()} records are waiting in queue\n'
        if pending:
            msg += f'{len(pending)} records to be send in the next batch'
            record = pending[-1]
        # if queue is empty and there is no pending records, message is effectively cleared
        info.set_status(msg, record)

    async def run_for(self, entity: DiscordHookEntity, client: HttpClient, info: TaskStatus):
        send_queue = self.queues[entity.name]

        bucket: Optional[str] = None
        until_next_try = 0
        to_be_sent: List[Record] = []

        while True:
            await asyncio.sleep(until_next_try)
            self.update_status(info, send_queue, to_be_sent)
            to_be_sent = await self.wait_for_records(send_queue, pending=to_be_sent)
            message, pending_records = self._prepare_message(to_be_sent, entity)
            self.update_status(info, send_queue, pending_records)
            if message is None:
                to_be_sent = pending_records
                continue

            request_details = self.endpoint.prepare(entity.url, message)
            response = await client.request_endpoint(self.logger, request_details)

            if isinstance(response, NoResponse):
                self.logger.warning(f'[{entity.name}] network error while sending message with Discord webhook, saving for the next try')
                until_next_try = 60
                continue  # implicitly saving messages in to_be_sent until the next try
            elif not response.ok:
                self.logger.warning(f'[{entity.name}] error while sending message with Discord webhook: got {response.status} ({response.reason or "No reason"}) {response.text}')
                self.logger.debug(f'raw message text:\n{message}')
                to_be_sent = pending_records # discard unsent messages as they might have been the cause of error
                until_next_try = 60
                continue
            if self._has_fatal_error(response, entity, message):
                break

            # if message got send successfully discard records except these that didn't fit into message
            if pending_records:
                self.logger.debug(f'carrying {len(pending_records)} pending records to the next loop')
            to_be_sent = pending_records

    @staticmethod
    async def wait_for_records(queue: asyncio.Queue, pending: List[Record]) -> List[Record]:
        to_be_sent = pending
        while len(to_be_sent) < DiscordEmbedLimits.EMBEDS_PER_MESSAGE:
            try:
                record = await asyncio.wait_for(queue.get(), 60 / DiscordEmbedLimits.EMBEDS_PER_MESSAGE)
                to_be_sent.append(record)
            except asyncio.TimeoutError:
                if to_be_sent:
                    break
        return to_be_sent

    def _prepare_message(self, to_be_sent: List[Record], entity: DiscordHookEntity) -> Tuple[Optional[dict], List[Record]]:
        try:
            message, pending_records = MessageFormatter.format(to_be_sent)
            if pending_records:
                self.logger.debug(f'[{entity.name}] out of {len(to_be_sent)} records {len(pending_records)} are left pending')
        except Exception as e:
            self.logger.exception(f'error happened while formatting message: {e}\nRaw records list: "{to_be_sent}"')
            return None, []
        try:
            limits_ok = MessageFormatter.check_limits(message)
        except Exception as e:
            self.logger.exception(f'error checking content length limits for message: {e}\nRaw message: "{message}"')
            return None, pending_records
        if not limits_ok:
            self.logger.warning(f'[{entity.name}] prepared message exceeded Discord length limits. Records will be discarded')
            self.logger.debug(f'[{entity.name}] message content:\n{message}')
            return None, pending_records

        return message, pending_records

    def _has_fatal_error(self, response: HttpResponse, entity: DiscordHookEntity, message: dict) -> bool:
        if response.ok:
            return False
        self.logger.warning(f'[{entity.name}] failed to send message: got {response.status} ({response.reason}) from {response.url}')
        if response.status == 400:
            self.logger.warning(f'[{entity.name}] message got rejected with {response.status} ({response.reason}), dropping it')
            self.logger.debug(f'[{entity.name}] request headers: {response.request_headers}')
            self.logger.debug(f'[{entity.name}] response headers: {response.headers}')
            self.logger.debug(f'[{entity.name}] raw request body: {json.dumps(message)}')
            return False
        elif response.status in [404, 401]:
            self.logger.warning(f'[{entity.name}] got {response.status} from webhook, interrupting operations. Check if webhook url is still valid: {response.url}')
            return True
        else:
            return False



import datetime
from abc import ABC
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

import dateutil.parser
from pydantic import Field, PositiveFloat

from avtdl.core.formatters import Fmt
from avtdl.core.interfaces import Record
from avtdl.core.monitors import PagedFeedMonitor, PagedFeedMonitorConfig, PagedFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.core.request import HttpClient
from avtdl.core.utils import JSONType, find_all


class IwaraPostRecord(Record, ABC):
    url: str
    """link to the post"""
    title: str
    """post title"""
    published: datetime.datetime
    """post publish date"""
    name: str
    """uploader visible name"""
    username: str
    """uploader username"""
    avatar_url: Optional[str] = None
    """link to the picture used as the uploader avatar"""
    private: bool
    """whether post is private or unlisted"""
    tags: List[str] = []
    """list of tags of the post"""

    def get_uid(self) -> str:
        return f'{self.url}'


@Plugins.register('iwr', Plugins.kind.ASSOCIATED_RECORD)
class IwaraVideoRecord(IwaraPostRecord):
    thumbnail_url: str
    """url of the video thumbnail"""
    preview_url: Optional[str]
    """url of the animated preview, if available"""
    size: int
    """video size, in bytes"""
    duration: int
    """video duration, in seconds"""

    def __str__(self) -> str:
        return f'{self.url}\n[{self.name}] {self.title}'

    def __repr__(self) -> str:
        template = '[{}] {} ({})'
        return template.format(self.name, self.title[:60], self.url)

    def as_embed(self) -> dict:
        profile_url = f'https://www.iwara.tv/profile/{self.username}'
        footer = '{}, {}'.format(Fmt.duration(self.duration), Fmt.size(self.size))
        return {
            'title': self.title,
            'url': self.url,
            'description': None,
            'image': {'url': self.thumbnail_url, '_preview': self.preview_url},
            'color': None,
            'author': {'name': self.name, 'url': profile_url, 'icon_url': self.avatar_url},
            'timestamp': self.published.isoformat(),
            'footer': {'text': footer},
            'fields': []
        }


@Plugins.register('iwr', Plugins.kind.ACTOR_CONFIG)
class IwaraMonitorConfig(PagedFeedMonitorConfig):
    pass


@Plugins.register('iwr', Plugins.kind.ACTOR_ENTITY)
class IwaraMonitorEntity(PagedFeedMonitorEntity):
    url: str = Field(default='', exclude=True)
    """url field is unused"""
    update_interval: PositiveFloat = 3600
    """how often update should happen, in seconds"""


@dataclass
class Context:
    page: int


class IsEmbeddedVideoError(Exception):
    """Raised when a parsed post embeds link to Youtube"""


@Plugins.register('iwr', Plugins.kind.ACTOR)
class IwaraMonitor(PagedFeedMonitor):
    """
    Monitor Iwara uploads

    Checks Latest Videos feed in reverse chronological order,
    does not retrieve descriptions and comments.
    Posts embedding videos from Youtube are skipped.
    Does not provide filtering capabilities, selecting records
    for further processing can be accomplished by using "filter.match" or other filters.
    """

    async def handle_first_page(self,
                                entity: IwaraMonitorEntity,
                                client: HttpClient) -> Tuple[Optional[Sequence[Record]], Optional[Context]]:
        data = await self._fetch_data(entity, client, None)
        records = self._parse_data(data)
        return records, Context(page=1)

    async def handle_next_page(self,
                               entity: IwaraMonitorEntity,
                               client: HttpClient,
                               context: Optional[Context]) -> Tuple[Optional[Sequence[Record]], Optional[Context]]:
        if context is None:
            return None, None
        data = await self._fetch_data(entity, client, context)
        records = self._parse_data(data)
        if records is not None:
            context.page += 1
        else:
            context = None
        return records, context

    async def _fetch_data(self,
                          entity: IwaraMonitorEntity,
                          client: HttpClient,
                          context: Optional[Context]) -> Optional[JSONType]:
        url = 'https://api.iwara.tv/videos'
        params = {'rating': 'all', 'sort': 'date'}
        if context:
            params['page'] = str(context.page)
        data = await self.request_json(url, entity, client, params=params)
        return data

    def _parse_data(self, data: JSONType) -> Optional[Sequence[IwaraVideoRecord]]:
        if data is None:
            return None
        if not isinstance(data, dict) or 'results' not in data:
            self.logger.warning(f'failed to parse API response: no results')
            self.logger.debug(f'raw response: {data}')
            return None
        results = data['results']
        records = []
        for result in results:
            try:
                record = self._parse_record(result)
            except IsEmbeddedVideoError as e:
                self.logger.debug(f'skipping embedded video: {e}')
            except Exception as e:
                self.logger.warning(f'failed to parse record: {type(e)}: {e}')
                self.logger.debug(f'raw item: {result}', exc_info=e)
            else:
                records.append(record)
        return records

    def _parse_record(self, data: JSONType) -> IwaraVideoRecord:
        if not isinstance(data, dict):
            raise ValueError(f'unexpected format')
        if 'user' not in data:
            raise ValueError('no user info')
        if data.get('embedUrl') is not None:
            raise IsEmbeddedVideoError(data['embedUrl'])
        if 'file' not in data:
            raise ValueError('no video info')

        user = data['user']
        if user.get('avatar') is None:
            avatar_url = None
        else:
            avatar_path = user['avatar']['name']
            avatar_id = user['avatar']['id']
            avatar_url = f'https://i.iwara.tv/image/avatar/{avatar_id}/{avatar_path}'

        tags = [str(tag) for tag in find_all(data, '$.tags[*].id')]

        file = data['file']
        file_id = file['id']
        thumbnail_id = int(data['thumbnail'])

        post_id = data['id']
        post_slug = data['slug'] or ''

        record = IwaraVideoRecord(
            url=f'https://www.iwara.tv/video/{post_id}/{post_slug}',
            title=data['title'],
            private=bool(data['private'] or data['unlisted'] or False),
            published=dateutil.parser.parse(data['createdAt']),
            tags=tags,
            name=user['name'],
            username=user['username'],
            avatar_url=avatar_url,
            duration=file.get('duration'),
            size=file['size'],
            thumbnail_url=f'https://i.iwara.tv/image/thumbnail/{file_id}/thumbnail-{thumbnail_id:02}.jpg',
            preview_url=f'https://i.iwara.tv/image/original/{file_id}/preview.webp'
        )
        return record



import abc
import datetime
import json
import logging
from abc import abstractmethod
from hashlib import sha1
from textwrap import shorten
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny, field_validator, model_validator

MAX_REPR_LEN = 60


def utcnow() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


class Record(BaseModel):
    '''Data entry, passed around from Monitors to Actions through Filters'''

    model_config = ConfigDict(use_attribute_docstrings=True)

    origin: Optional[str] = Field(default=None, exclude=True)
    """semicolon-separated names of actor and entity record originated from"""
    chain: str = Field(default='', exclude=True)
    """name of the Chain this record is going through.
    Empty string means it was just produced and should go to every subscriber"""
    created_at: datetime.datetime = Field(default_factory=utcnow, exclude=True)
    """record creation timestamp"""
    class_name: str = Field(default='', validate_default=True, exclude=True)
    """class name of specific Record implementation, used for deserialization"""

    @field_validator('class_name')
    @classmethod
    def set_class_name(cls, _: str) -> str:
        return cls.__name__

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        all_known_record_types[cls.__name__] = cls

    @abstractmethod
    def __str__(self) -> str:
        '''Text representation of the record to be sent in a message, written to a file etc.'''

    @abstractmethod
    def __repr__(self) -> str:
        '''Short text representation of the record to be printed in logs'''

    def __eq__(self, other) -> bool:
        if not isinstance(other, Record):
            return NotImplemented
        return self.model_dump() == other.model_dump()

    def get_uid(self) -> str:
        '''A string that is the same for different versions of the same record'''
        return self.hash()

    def as_timezone(self, timezone: Optional[datetime.tzinfo] = None) -> 'Record':
        fields = dict(self)
        for k, v in fields.items():
            if isinstance(v, Record):
                fields[k] = v.as_timezone(timezone)
            if isinstance(v, datetime.datetime):
                fields[k] = v.astimezone(timezone)
        record_copy = self.model_validate(fields)
        return record_copy

    def as_json(self, indent: Union[int, str, None] = None) -> str:
        return json.dumps(self.model_dump(), sort_keys=True, ensure_ascii=False, default=str, indent=indent)

    def as_embed(self) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        embed_title_max_length = 256
        embed_description_max_length = 4096

        text = str(self)
        if text.find('\n') > -1:
            title, description = text.split('\n', 1)
        else:
            title, description = '', text
        title = shorten(title, embed_title_max_length)
        description = shorten(description, embed_description_max_length)
        return {'title': title, 'description': description}

    def hash(self) -> str:
        record_hash = sha1(self.as_json().encode())
        return record_hash.hexdigest()


all_known_record_types: Dict[str, type[Record]] = {}


def get_record_type(name: str) -> Optional[type[Record]]:
    """return Record descendant class with given class name"""
    return all_known_record_types.get(name)


class TextRecord(Record):
    """
    Simplest record, containing only a single text field
    """

    text: str
    """content of the record"""

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f'TextRecord("{shorten(self.text, MAX_REPR_LEN)}")'


class EventType:
    generic: str = 'generic'
    error: str = 'error'
    started: str = 'started'
    finished: str = 'finished'


class Event(Record):
    """
    Record produced by an internal event (usually error) inside the plugin
    """

    event_type: str = EventType.generic
    """text describing the nature of event, can be used to filter classes of events, such as errors"""
    text: str
    """text describing specific even details"""
    record: SerializeAsAny[Optional[Record]] = Field(exclude=True, default=None)
    """record that was being processed when this event happened"""

    def __str__(self):
        return self.text

    def __repr__(self):
        text = shorten(self.text, MAX_REPR_LEN)
        return f'Event(event_type="{self.event_type}", text="{text}")'

    def model_post_init(self, __context):
        if self.record is not None:
            self.origin = self.record.origin
            self.chain = self.record.chain


class OpaqueRecord(Record):
    """Record without predefined fields and structure"""
    model_config = ConfigDict(extra='allow')

    @model_validator(mode='before')
    @classmethod
    def check_fields_overwrite(cls, data: Any) -> Any:
        if isinstance(data, dict):
            for name, value in data.items():
                if name in cls.model_fields:
                    logger = logging.getLogger('OpaqueRecord')
                    logger.warning(f'internal field "{name}" was overwritten by user-defined value "{value}"')
        return data

    def __str__(self) -> str:
        return self.model_dump_json(indent=4)

    def __repr__(self) -> str:
        fields = ', '.join((f'{field}={value}' for field, value in self.model_dump().items()))
        return f'{self.__class__.__name__}({fields})'


class AbstractRecordsStorage(abc.ABC):
    """Interface for accessing persistent records storage from web ui"""

    @abstractmethod
    def feeds(self) -> List[Tuple[str, int]]:
        """return names and number of records of distinct feeds storage currently has"""

    @abstractmethod
    def page_count(self, per_page: int, feed: Optional[str] = None) -> int:
        """return total number of pages"""

    @abstractmethod
    def load_page(self, page: Optional[int], per_page: int, desc: bool = True, feed: Optional[str] = None) -> List[
        Record]:
        """return content of specific page as a list of Record instances"""



import datetime
import glob
import hashlib
import logging
import os
import re
import shutil
import urllib.parse
from pathlib import Path
from typing import Any, Iterable, List, Optional

from avtdl.core.download import download_file
from avtdl.core.formatters import sanitize_filename
from avtdl.core.interfaces import Record
from avtdl.core.request import HttpClient


def find_file(path: Path) -> List[Path]:
    """given path to a file without extension, return list of existing files with exact path and name and any extension"""
    if not path.parent.exists():
        return []
    try:
        return [p for p in path.parent.iterdir() if p.stem == path.name]
    except OSError:
        return []


def find_with_suffix(path: Path, suffix_template: str) -> List[Path]:
    """given path to possible existing file name without extension and a suffix template,
    find and return a sorted list of all existing files with given name and suffix.
    The suffix_template parameter must contain "{i}" exactly once"""
    if suffix_template.count("{i}") != 1:
        raise ValueError("The suffix_template must contain '{i}' exactly once.")
    if not path.parent.exists():
        return []
    base_name = strip_rename_suffix(path.name, suffix_template)
    pattern = f'{glob.escape(base_name)}*'
    matches = []
    matches_with_suffix = []
    for p in path.parent.glob(pattern):
        if p.stem == base_name:
            matches.append(p)
        elif strip_rename_suffix(p.stem, suffix_template) == base_name:
            matches_with_suffix.append(p)

    matches_with_suffix.sort()
    matches.extend(matches_with_suffix)
    return matches


def strip_rename_suffix(name: str, suffix_template: str) -> str:
    """given file name without extension, remove suffix matching suffix_template"""
    re_suffix = re.escape(suffix_template).replace(re.escape('{i}'), r'\d+') + '$'
    new_name = re.sub(re_suffix, '', name)
    return new_name


def find_free_suffix(path: Path, suffix_template: str) -> Path:
    """given path to possible existing file name without extension, find a free name
    generated by appending suffix_template to original filename, such that no file exists
    with that name and any extension. The suffix_template parameter must contain "{i}" exactly once"""
    if suffix_template.count('{i}') != 1:
        raise ValueError("The suffix_template must contain '{i}' exactly once.")

    base_name = path.name
    new_path = Path(path)

    i = 0
    while find_file(new_path):
        i += 1
        suffix = suffix_template.format(i=i)
        new_path = path.with_name(f'{base_name}{suffix}')
    return new_path


def has_expired(file: Path, ttl: Optional[float]) -> bool:
    """check if file was created earlier than ttl hours ago"""
    if ttl is None:
        return False
    if ttl == 0:
        return True
    if not file.exists():
        return True
    try:
        stat = file.stat()
    except OSError:
        return True
    modified = datetime.datetime.fromtimestamp(stat.st_ctime)
    now = datetime.datetime.now()
    hours_passed = (now - modified).total_seconds() / 3600

    return hours_passed > ttl


class FileCache:
    """
    Provide interface to store and retrieve url content locally

    Uses name of the record field containing the url, url host
    and md5(url) to generate unique path for an url.

    cache_directory: path to cache directory. Multiple instances
    pointing to the same directory should be able to coexist
    (at least read-only and write-only ones)

    partial_file_suffix: new file initially has this extension and is then
    renamed once actual extension becomes known after download

    This plugin is under active development. The way it works might change,
    options might be renamed or removed.
    """

    RENAME_SUFFIX = ' [{i}]'

    def __init__(self, cache_directory: Path, partial_file_suffix: str):
        self.cache_directory = cache_directory.resolve()
        self.partial_file_suffix = partial_file_suffix
        self.logger = logging.getLogger('cache')

    @staticmethod
    def _field_name_by_value(record: Record, value: Any, default='default') -> str:
        """return name of record's field containing value, or default if it wasn't found"""
        fields = type(record).model_fields.keys()
        for field in fields:
            field_value = getattr(record, field, None)
            if field_value == value:
                return field
            if isinstance(field_value, Iterable) and value in field_value:
                return field
        return default

    def filename_for(self, record: Record, url: str) -> Path:
        """generate path and name (without extension) for given url"""
        subdirectory = Path(urllib.parse.urlparse(url).netloc) / self._field_name_by_value(record, url)
        url_hash = hashlib.md5(url.encode()).hexdigest().upper()
        name = sanitize_filename(f'{url_hash}', collapse=True)
        name = Path(name)
        filename = self.cache_directory / subdirectory / name
        return filename

    def retrieve(self, record: Record, url: str) -> Optional[Path]:
        path = self.filename_for(record, url)
        file = self._find_file(path, url)
        return file

    async def store(self, logger: logging.Logger, client: HttpClient,
                    record: Record, url: str,
                    replace_after: Optional[float] = None, external_path: Optional[Path] = None) -> Optional[Path]:
        """download and store or find existing local copy of the url, return path to the local file.
        Use existing file "unless replace_after" hours has passed since it was created, otherwise
        pick a new name and download anyway. If external_path is provided and is a file, it is
        copied to cache directory instead of downloading"""
        store_path = self.filename_for(record, url)
        file = self._find_file(store_path, url)
        if file and not has_expired(file, replace_after):
            self.logger.debug(f'reusing stored file "{file}" for "{url}"')
            return file
        store_path = find_free_suffix(store_path, self.RENAME_SUFFIX)
        try:
            store_path.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logger.warning(f'failed to create directory hierarchy "{store_path.parent}", skipping record {record!r}')
            return None
        if external_path and external_path.exists() and external_path.is_file():
            logger.debug(f'reusing external file "{external_path}" for "{url}"')
            try:
                shutil.copy2(external_path, store_path)
                return store_path
            except OSError as e:
                logger.warning(f'failed to copy external file "{external_path}" to "{store_path}", downloading')
        return await self._download_file(logger, client, url, store_path)

    async def _download_file(self, logger: logging.Logger, client: HttpClient, url: str, path: Path) -> Optional[Path]:
        """download file from given url, store to path, return path if successful"""
        temp_path = path.with_name(path.name + self.partial_file_suffix)
        info = await download_file(url, temp_path, client.session, logger=logger)
        if info is None:
            return None
        final_path = temp_path.with_suffix(info.extension)
        try:
            logger.debug(f'moving "{temp_path}" to "{final_path}"')
            os.replace(temp_path, final_path)
            return final_path
        except Exception as e:
            message = f'failed to move file "{temp_path}" to desired location "{final_path}": {e}'
            logger.warning(message)
            return None

    def _find_file(self, path: Path, url: str) -> Optional[Path]:
        """given path to a file without extension, find and return existing file"""
        files = find_with_suffix(path, self.RENAME_SUFFIX)
        if len(files) == 1:
            file = files[0]
            self.logger.debug(f'url "{url} is stored at "{file}"')
            return file
        elif len(files) > 1:
            for file in reversed(files):
                if url.endswith(file.suffix):
                    self.logger.debug(f'url "{url} is stored at "{file}" (picked best match out of {len(files)})')
                    return file
            self.logger.warning(f'found multiple files suitable for url "{url}": {files}')
            return files[0]
        else:
            return None
